from __future__ import annotations

import abc
import requests
from typing import Any, Coroutine, TYPE_CHECKING

import aiohttp
import requests.structures

from ..core.param import RequestType
from ..core.exception import (
    RequestTypeNotSupported, 
    AuthorizationAPIKeyError, 
    AsyncClientAPISessionNotAvailable
)

if TYPE_CHECKING:
    from ..core.interaction import Interaction

class APIClientInterface(metaclass = abc.ABCMeta):
    """
        The following abstract class defines a general Client API. 
        A client is used in order to connect and interact with an external API.

        The key method of an API client is the `api` function that is used 
        to execute the actual requests.
    """
    @classmethod
    def __subclasshook__(cls, subclass):
        return (
                hasattr(subclass, 'api') and callable(subclass.api)
            or NotImplemented
        )

    @abc.abstractmethod
    def api(self, type: str, url: str, *args: tuple, **kwargs: dict[str, Any]) -> requests.Response  | Coroutine[None, None, requests.Response]:
        """
            Function in charge to execute a request to an API endpoint.

            Parameters:
                type: request's type ([`RequestType`][cyhole.core.param.RequestType]).
                url: API endpoint.
                args: additional input parameters to provided.
                kwargs: additional input parameters to provided.

            Returns:
                The response object structured as `request` library. Observe that the return type could be a `Coroutine` in case of implementation  of an `async` client.

            Raises:
                RequestTypeNotSupported: if the request type is not a valid value of [`RequestType`][cyhole.core.param.RequestType].
                AuthorizationAPIKeyError: if the response return a 401 code.
        """
        raise NotImplementedError

class APIClient(APIClientInterface):
    """
        This client is designed to manage all the REST calls to an external API by using `requests` module.

        !!! info
            As a conseguence of using `requests` module, this client can be used to perform 
            the API requests in **synchronous** logic and achieve the parallelism by using threads.

        Use this class as middlelayer to manage all the requests to an external API. 
        By default, all new `Interaction` should have the synchronous client that inherits from this class.

        During the creation of the object is possible to specify some global configurations.

        Parameters:
            headers: headers used globally in all API requests.
    """
    def __init__(self, interaction: Interaction, headers: Any | None = None) -> None:
        self._interaction = interaction
        self.headers = headers
        return

    def api(self, type: str, url: str, *args: tuple, **kwargs: dict[str, Any]) -> requests.Response:

        # check for headers
        if self.headers:
            kwargs["headers"] = self.headers

        # execute request
        match type:
            case RequestType.GET.value:
                response = requests.get(url, *args, **kwargs)
            case RequestType.POST.value:
                response = requests.post(url, *args, **kwargs)
            case _:
                raise RequestTypeNotSupported(f"Request '{type}' not supported.")

        # check response
        match response.status_code:
            case 401:
                raise AuthorizationAPIKeyError("Check API Key")
            case _:
                response.raise_for_status()
                return response

class AsyncAPIClient(APIClientInterface):
    """
        This client is designed to manage all the REST calls to an external API by using `aiohttp` module.

        !!! info
            As a conseguence of using `aiohttp` module, this client can be used to perform 
            the API requests in **asynchronous** logic of `async` paradigm.

        Use this class as middlelayer to manage all the requests to an external API. 
        By default, all new `Interaction` should have the asynchronous client that inherits from this class.

        During the creation of the object is possible to specify some global configurations.

        Parameters:
            headers: headers used globally in all API requests.
    """
    def __init__(self, interaction: Interaction, headers: Any | None = None) -> None:
        self._session: aiohttp.ClientSession | None = None
        self._interaction = interaction
        self.headers = headers
        return

    async def __aenter__(self):
        """Open a new session."""
        self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, exc_traceback):
        """Exits from the session."""
        await self.close()
        return

    def is_connected(self) -> bool:
        """Check if the session is available."""
        if self._session is None:
            return False
        return not self._session.closed

    def connect(self) -> None:
        """Init a new session."""
        if self._session is None:
            self._session = aiohttp.ClientSession()
        return

    async def close(self) -> None:
        """Close current available session."""

        if self._session is None:
            raise AsyncClientAPISessionNotAvailable("No session currently available.")

        # close and reset
        await self._session.close()
        self._session = None
        return

    async def api(self, type: str, url: str, *args: tuple, **kwargs: dict[str, Any]) -> requests.Response:

        # check session availability
        if self._session is None:
            raise AsyncClientAPISessionNotAvailable("No session currently available.")

        # check for headers
        if self.headers:
            kwargs["headers"] = self.headers

        # clean params
        if "params" in kwargs:
            kwargs["params"] = self._clean_params(kwargs["params"].copy())

        # execute request
        match type:
            case RequestType.GET.value:
                async with self._session.get(url, *args, **kwargs) as resp:
                    response = await self._to_requests_response(resp)
            case RequestType.POST.value:
                async with self._session.post(url, *args, **kwargs) as resp:
                    response = await self._to_requests_response(resp)
            case _:
                raise RequestTypeNotSupported(f"Request '{type}' not supported.")

        # check response
        match response.status_code:
            case 401:
                raise AuthorizationAPIKeyError("Check API Key")
            case _:
                response.raise_for_status()
                return response

    async def _to_requests_response(self, response: aiohttp.ClientResponse) -> requests.Response:
        """
            This internal function is used to convert a response obtained 
            from `aiohttp` to the one from `requests` library.

            Parameters:
                response: response from `aiohttp`.

            Return:
                response from `requests`.
        """
        _response = requests.Response()
        _response.url = str(response.url)
        _response.headers = requests.structures.CaseInsensitiveDict(response.headers)
        _response.status_code = response.status
        _response.reason = response.reason if response.reason else "Not Found"
        _response.encoding = response.get_encoding()
        _response._content = await response.read()

        return _response

    def _clean_params(self, params: dict[str, Any]) -> dict[str, Any]:
        """
            The `aiohttp` requires to remove the keys with `None` value. 
            This function takes as input the original `params` and returns a new 
            object removed from `None`keys.

            arameters:
                params: original params.

            Return:
                new params removed from the keys with `None`valiues.
        """
        return {key: value for key, value in params.items() if value is not None}


from __future__ import annotations
from typing import TYPE_CHECKING, Any

from ..core.client import APIClient, AsyncAPIClient
from ..jupiter.schema import (
    GetPriceResponse,
    GetQuoteInput,
    GetQuoteResponse,
    GetQuoteTokensResponse,
    GetQuoteProgramIdLabelResponse,
    PostSwapBody,
    PostSwapResponse,
    GetTokenListResponse,
    PostLimitOrderCreateBody,
    PostLimitOrderCreateResponse,
    PostLimitOrderCancelBody,
    PostLimitOrderCancelResponse,
    GetLimitOrderOpenResponse,
    GetLimitOrderHistoryResponse,
    GetLimitOrderTradeHistoryResponse
)
from ..jupiter.param import JupiterTokenListType

if TYPE_CHECKING:
    from ..jupiter.interaction import Jupiter

class JupiterClient(APIClient):
    """
        Client used for synchronous API calls for `Jupiter` interaction.
    """

    def __init__(self, interaction: Jupiter, headers: Any | None = None) -> None:
        super().__init__(interaction, headers)
        self._interaction: Jupiter = self._interaction

    def get_price(self, address: list[str], vs_address: str | None = None) -> GetPriceResponse:
        """
            Call the Jupiter's **[Price](https://station.jup.ag/docs/apis/price-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_price`][cyhole.jupiter.interaction.Jupiter._get_price].
        """
        return self._interaction._get_price(True, address, vs_address)

    def get_quote(self, input: GetQuoteInput) -> GetQuoteResponse:
        """
            Call the Jupiter's **[Get Quote](https://station.jup.ag/api-v6/get-quote)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote`][cyhole.jupiter.interaction.Jupiter._get_quote].
        """
        return self._interaction._get_quote(True, input)

    def get_quote_tokens(self) -> GetQuoteTokensResponse:
        """
            Call the Jupiter's **[Get Quote Tokens](https://station.jup.ag/api-v6/get-tokens)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote_tokens`][cyhole.jupiter.interaction.Jupiter._get_quote_tokens].
        """
        return self._interaction._get_quote_tokens(True)

    def get_quote_program_id_label(self) -> GetQuoteProgramIdLabelResponse:
        """
            Call the Jupiter's **[Get Quote Program ID to Label](https://station.jup.ag/api-v6/get-program-id-to-label)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote_program_id_label`][cyhole.jupiter.interaction.Jupiter._get_quote_program_id_label].
        """
        return self._interaction._get_quote_program_id_label(True)

    def post_swap(self, body: PostSwapBody) -> PostSwapResponse:
        """
            Call the Jupiter's **[Post Swap](https://station.jup.ag/api-v6/post-swap)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._post_swap`][cyhole.jupiter.interaction.Jupiter._post_swap].
        """
        return self._interaction._post_swap(True, body)

    def get_token_list(self, type: str = JupiterTokenListType.STRICT.value, banned: None | bool = None) -> GetTokenListResponse:
        """
            Call the Jupiter's **[Token List](https://station.jup.ag/docs/token-list/token-list-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_token_list`][cyhole.jupiter.interaction.Jupiter._get_token_list].
        """
        return self._interaction._get_token_list(True, type, banned)

    def post_limit_order_create(self, body: PostLimitOrderCreateBody) -> PostLimitOrderCreateResponse:
        """
            Call the Jupiter's **[Post Limit Order - Create](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._post_limit_order_create`][cyhole.jupiter.interaction.Jupiter._post_limit_order_create].
        """
        return self._interaction._post_limit_order_create(True, body)

    def post_limit_order_cancel(self, body: PostLimitOrderCancelBody) -> PostLimitOrderCancelResponse:
        """
            Call the Jupiter's **[Post Limit Order - Cancel](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._post_limit_order_cancel`][cyhole.jupiter.interaction.Jupiter._post_limit_order_cancel].
        """
        return self._interaction._post_limit_order_cancel(True, body)

    def get_limit_order_open(
            self,
            wallet: str | None = None,
            input_token: str | None = None,
            output_token: str | None = None
        ) -> GetLimitOrderOpenResponse:
        """
            Call the Jupiter's **[Get Limit Order - Open](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_limit_order_open`][cyhole.jupiter.interaction.Jupiter._get_limit_order_open].
        """
        return self._interaction._get_limit_order_open(True, wallet, input_token, output_token)

    def get_limit_order_history(
        self,
        wallet: str,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderHistoryResponse:
        """
            Call the Jupiter's **[Get Limit Order - History](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_limit_order_history`][cyhole.jupiter.interaction.Jupiter._get_limit_order_history].
        """
        return self._interaction._get_limit_order_history(True, wallet, cursor, skip, take)

    def get_limit_order_trade_history(
        self,
        wallet: str | None = None,
        input_token: str | None = None,
        output_token: str | None = None,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderTradeHistoryResponse:
        """
            Call the Jupiter's **[Get Limit Order - Trade History](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_limit_order_trade_history`][cyhole.jupiter.interaction.Jupiter._get_limit_order_trade_history].
        """
        return self._interaction._get_limit_order_trade_history(True, wallet, input_token, output_token, cursor, skip, take)

class JupiterAsyncClient(AsyncAPIClient):
    """
        Client used for asynchronous API calls for `Jupiter` interaction.
    """

    def __init__(self, interaction: Jupiter, headers: Any | None = None) -> None:
        super().__init__(interaction, headers)
        self._interaction: Jupiter = self._interaction

    async def get_price(self, address: list[str], vs_address: str | None = None) -> GetPriceResponse:
        """
            Call the Jupiter's **[Price](https://station.jup.ag/docs/apis/price-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_price`][cyhole.jupiter.interaction.Jupiter._get_price].
        """
        return await self._interaction._get_price(False, address, vs_address)

    async def get_quote(self, input: GetQuoteInput) -> GetQuoteResponse:
        """
            Call the Jupiter's **[Get Quote](https://station.jup.ag/api-v6/get-quote)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote`][cyhole.jupiter.interaction.Jupiter._get_quote].
        """
        return await self._interaction._get_quote(False, input)

    async def get_quote_tokens(self) -> GetQuoteTokensResponse:
        """
            Call the Jupiter's **[Get Quote Tokens](https://station.jup.ag/api-v6/get-tokens)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote_tokens`][cyhole.jupiter.interaction.Jupiter._get_quote_tokens].
        """
        return await self._interaction._get_quote_tokens(False)

    async def get_quote_program_id_label(self) -> GetQuoteProgramIdLabelResponse:
        """
            Call the Jupiter's **[Get Quote Program ID to Label](https://station.jup.ag/api-v6/get-program-id-to-label)** API endpoint for synchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote_program_id_label`][cyhole.jupiter.interaction.Jupiter._get_quote_program_id_label].
        """
        return await self._interaction._get_quote_program_id_label(False)

    async def post_swap(self, body: PostSwapBody) -> PostSwapResponse:
        """
            Call the Jupiter's **[Post Swap](https://station.jup.ag/api-v6/post-swap)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_quote_program_id_label`][cyhole.jupiter.interaction.Jupiter._get_quote_program_id_label].
        """
        return await self._interaction._post_swap(False, body)

    async def get_token_list(self, type: str = JupiterTokenListType.STRICT.value, banned: None | bool = None) -> GetTokenListResponse:
        """
            Call the Jupiter's **[Token List](https://station.jup.ag/docs/token-list/token-list-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_token_list`][cyhole.jupiter.interaction.Jupiter._get_token_list].
        """
        return await self._interaction._get_token_list(False, type, banned)

    async def post_limit_order_create(self, body: PostLimitOrderCreateBody) -> PostLimitOrderCreateResponse:
        """
            Call the Jupiter's **[Post Limit Order - Create](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._post_limit_order_create`][cyhole.jupiter.interaction.Jupiter._post_limit_order_create].
        """
        return await self._interaction._post_limit_order_create(False, body)

    async def post_limit_order_cancel(self, body: PostLimitOrderCancelBody) -> PostLimitOrderCancelResponse:
        """
            Call the Jupiter's **[Post Limit Order - Cancel](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._post_limit_order_cancel`][cyhole.jupiter.interaction.Jupiter._post_limit_order_cancel].
        """
        return await self._interaction._post_limit_order_cancel(False, body)

    async def get_limit_order_open(
            self,
            wallet: str | None = None,
            input_token: str | None = None,
            output_token: str | None = None
        ) -> GetLimitOrderOpenResponse:
        """
            Call the Jupiter's **[Get Limit Order - Open](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_limit_order_open`][cyhole.jupiter.interaction.Jupiter._get_limit_order_open].
        """
        return await self._interaction._get_limit_order_open(False, wallet, input_token, output_token)

    async def get_limit_order_history(
        self,
        wallet: str,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderHistoryResponse:
        """
            Call the Jupiter's **[Get Limit Order - History](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_limit_order_history`][cyhole.jupiter.interaction.Jupiter._get_limit_order_history].
        """
        return await self._interaction._get_limit_order_history(False, wallet, cursor, skip, take)

    async def get_limit_order_trade_history(
        self,
        wallet: str | None = None,
        input_token: str | None = None,
        output_token: str | None = None,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderTradeHistoryResponse:
        """
            Call the Jupiter's **[Get Limit Order - Trade History](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint for asynchronous logic. 
            All the API endopint details are available on [`Jupiter._get_limit_order_trade_history`][cyhole.jupiter.interaction.Jupiter._get_limit_order_trade_history].
        """
        return await self._interaction._get_limit_order_trade_history(False, wallet, input_token, output_token, cursor, skip, take)


from __future__ import annotations
import requests
from datetime import datetime
from typing import TYPE_CHECKING, Any

from ..core.client import APIClient, AsyncAPIClient
from ..core.exception import AuthorizationAPIKeyError
from ..birdeye.exception import BirdeyeAuthorisationError
from ..birdeye.param import (
    BirdeyeOrder,
    BirdeyeSort,
    BirdeyeTradeType
)
from ..birdeye.schema import (
    GetTokenListResponse,
    GetTokenSecurityResponse,
    GetTokenCreationInfoResponse,
    GetTokenOverviewResponse,
    GetPriceResponse,
    GetPriceMultipleResponse,
    GetPriceHistoricalResponse,
    GetTradesTokenResponse,
    GetTradesPairResponse,
    GetOHLCVTokenPairResponse,
    GetOHLCVBaseQuoteResponse,
    GetWalletSupportedNetworksResponse
)

if TYPE_CHECKING:
    from ..birdeye.interaction import Birdeye

class BirdeyeClient(APIClient):
    """
        Client used for synchronous API calls for `Birdeye` interaction.
    """

    def __init__(self, interaction: Birdeye, headers: Any | None = None) -> None:
        super().__init__(interaction, headers)
        self._interaction: Birdeye = self._interaction

    def api(self, type: str, url: str, *args: tuple, **kwargs: dict[str, Any]) -> requests.Response:
        # overide function to manage client specific exceptions
        try:
            return super().api(type, url, *args, **kwargs)
        except AuthorizationAPIKeyError:
            raise BirdeyeAuthorisationError

    def get_token_list(self, sort_by: str = BirdeyeSort.SORT_V24HUSD.value, order_by: str = BirdeyeOrder.DESCENDING.value, offset: int | None = None, limit: int | None = None) -> GetTokenListResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Token - List](https://docs.birdeye.so/reference/get_defi-tokenlist)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_list`][cyhole.birdeye.interaction.Birdeye._get_token_list].
        """
        return self._interaction._get_token_list(True, sort_by, order_by, offset, limit)

    def get_token_creation_info(self, address: str) -> GetTokenCreationInfoResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Token - Creation Token Info](https://docs.birdeye.so/reference/get_defi-token-creation-info)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_creation_info`][cyhole.birdeye.interaction.Birdeye._get_token_creation_info].
        """
        return self._interaction._get_token_creation_info(True, address)

    def get_token_security(self, address: str) -> GetTokenSecurityResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Token - Security](https://docs.birdeye.so/reference/get_defi-token-security)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_security`][cyhole.birdeye.interaction.Birdeye._get_token_security].
        """
        return self._interaction._get_token_security(True, address)

    def get_token_overview(self, address: str) -> GetTokenOverviewResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Token - Overview](https://docs.birdeye.so/reference/get_defi-token-overview)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_overview`][cyhole.birdeye.interaction.Birdeye._get_token_overview].
        """
        return self._interaction._get_token_overview(True, address)

    def get_price(self, address: str, include_liquidity: bool | None = None) -> GetPriceResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Price](https://docs.birdeye.so/reference/get_defi-price)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_price`][cyhole.birdeye.interaction.Birdeye._get_price].
        """
        return self._interaction._get_price(True, address, include_liquidity)

    def get_price_multiple(self, list_address: list[str], include_liquidity: bool | None = None) -> GetPriceMultipleResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Price - Multiple](https://docs.birdeye.so/reference/get_defi-multi-price)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_price_multiple`][cyhole.birdeye.interaction.Birdeye._get_price_multiple].
        """
        return self._interaction._get_price_multiple(True, list_address, include_liquidity)

    def get_price_historical(self, address: str, address_type: str, timeframe: str, dt_from: datetime, dt_to: datetime | None = None) -> GetPriceHistoricalResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Price - Historical](https://docs.birdeye.so/reference/get_defi-history-price)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_price_historical`][cyhole.birdeye.interaction.Birdeye._get_price_historical].
        """
        return self._interaction._get_price_historical(True, address, address_type, timeframe, dt_from, dt_to)

    def get_trades_token(self, address: str, trade_type: str = BirdeyeTradeType.SWAP.value, offset: int | None = None, limit: int | None = None) -> GetTradesTokenResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Trades - Token](https://docs.birdeye.so/reference/get_defi-txs-token)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_trades_token`][cyhole.birdeye.interaction.Birdeye._get_trades_token].
        """
        return self._interaction._get_trades_token(True, address, trade_type, offset, limit)

    def get_trades_pair(self, address: str, trade_type: str = BirdeyeTradeType.SWAP.value, order_by: str = BirdeyeOrder.DESCENDING.value, offset: int | None = None, limit: int | None = None) -> GetTradesPairResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Trades - Pair](https://docs.birdeye.so/reference/get_defi-txs-pair)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_trades_pair`][cyhole.birdeye.interaction.Birdeye._get_trades_pair].
        """
        return self._interaction._get_trades_pair(True, address, trade_type, order_by, offset, limit)

    def get_ohlcv(self, address: str, address_type: str, timeframe: str, dt_from: datetime, dt_to: datetime | None = None) -> GetOHLCVTokenPairResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[OHLCV - Token/Pair](https://docs.birdeye.so/reference/get_defi-ohlcv)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_ohlcv`][cyhole.birdeye.interaction.Birdeye._get_ohlcv].
        """
        return self._interaction._get_ohlcv(True, address, address_type, timeframe, dt_from, dt_to)

    def get_ohlcv_base_quote(self, base_address: str, quote_address: str, timeframe: str, dt_from: datetime, dt_to: datetime | None = None) -> GetOHLCVBaseQuoteResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[OHLCV - Base/Quote](https://docs.birdeye.so/reference/get_defi-ohlcv-base-quote)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_ohlcv_base_quote`][cyhole.birdeye.interaction.Birdeye._get_ohlcv_base_quote].
        """
        return self._interaction._get_ohlcv_base_quote(True, base_address, quote_address, timeframe, dt_from, dt_to)

    def get_wallet_supported_networks(self) -> GetWalletSupportedNetworksResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Wallet - Supported Networks](https://docs.birdeye.so/reference/get_v1-wallet-list-supported-chain)** for synchronous logic. 
            All the API endopint details are available on [`Birdeye._get_wallet_supported_networks`][cyhole.birdeye.interaction.Birdeye._get_wallet_supported_networks].
        """
        return self._interaction._get_wallet_supported_networks(True)

class BirdeyeAsyncClient(AsyncAPIClient):
    """
        Client used for asynchronous API calls for `Birdeye` interaction.
    """

    def __init__(self, interaction: Birdeye, headers: Any | None = None) -> None:
        super().__init__(interaction, headers)
        self._interaction: Birdeye = self._interaction

    async def api(self, type: str, url: str, *args: tuple, **kwargs: dict[str, Any]) -> requests.Response:
        # overide function to manage client specific exceptions
        try:
            return await super().api(type, url, *args, **kwargs)
        except AuthorizationAPIKeyError:
            raise BirdeyeAuthorisationError

    async def get_token_list(self, sort_by: str = BirdeyeSort.SORT_V24HUSD.value, order_by: str = BirdeyeOrder.DESCENDING.value, offset: int | None = None, limit: int | None = None
    ) -> GetTokenListResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Token - List](https://docs.birdeye.so/reference/get_defi-tokenlist)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_list`][cyhole.birdeye.interaction.Birdeye._get_token_list].
        """
        return await self._interaction._get_token_list(False, sort_by, order_by, offset, limit)

    async def get_token_creation_info(self, address: str) -> GetTokenCreationInfoResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Token - Creation Token Info](https://docs.birdeye.so/reference/get_defi-token-creation-info)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_creation_info`][cyhole.birdeye.interaction.Birdeye._get_token_creation_info].
        """
        return await self._interaction._get_token_creation_info(False, address)

    async def get_token_security(self, address: str) -> GetTokenSecurityResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Token - Security](https://docs.birdeye.so/reference/get_defi-token-security)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_security`][cyhole.birdeye.interaction.Birdeye._get_token_security].
        """
        return await self._interaction._get_token_security(False, address)

    async def get_token_overview(self, address: str) -> GetTokenOverviewResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Token - Overview](https://docs.birdeye.so/reference/get_defi-token-overview)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_token_overview`][cyhole.birdeye.interaction.Birdeye._get_token_overview].
        """
        return await self._interaction._get_token_overview(False, address)

    async def get_price(self, address: str, include_liquidity: bool | None = None) -> GetPriceResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Price](https://docs.birdeye.so/reference/get_defi-price)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_price`][cyhole.birdeye.interaction.Birdeye._get_price].
        """
        return await self._interaction._get_price(False, address, include_liquidity)

    async def get_price_multiple(self, list_address: list[str], include_liquidity: bool | None = None) -> GetPriceMultipleResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Price - Multiple](https://docs.birdeye.so/reference/get_defi-multi-price)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_price_multiple`][cyhole.birdeye.interaction.Birdeye._get_price_multiple].
        """
        return await self._interaction._get_price_multiple(False, list_address, include_liquidity)

    async def get_price_historical(self, address: str, address_type: str, timeframe: str, dt_from: datetime, dt_to: datetime | None = None) -> GetPriceHistoricalResponse:
        """
            Call the Birdeye's **PUBLIC** API endpoint **[Price - Historical](https://docs.birdeye.so/reference/get_defi-history-price)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_price_historical`][cyhole.birdeye.interaction.Birdeye._get_price_historical].
        """
        return await self._interaction._get_price_historical(False, address, address_type, timeframe, dt_from, dt_to)

    async def get_trades_token(self, address: str, trade_type: str = BirdeyeTradeType.SWAP.value, offset: int | None = None, limit: int | None = None) -> GetTradesTokenResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Trades - Token](https://docs.birdeye.so/reference/get_defi-txs-token)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_trades_token`][cyhole.birdeye.interaction.Birdeye._get_trades_token].
        """
        return await self._interaction._get_trades_token(False, address, trade_type, offset, limit)

    async def get_trades_pair(self, address: str, trade_type: str = BirdeyeTradeType.SWAP.value, order_by: str = BirdeyeOrder.DESCENDING.value, offset: int | None = None, limit: int | None = None) -> GetTradesPairResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Trades - Pair](https://docs.birdeye.so/reference/get_defi-txs-pair)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_trades_pair`][cyhole.birdeye.interaction.Birdeye._get_trades_pair].
        """
        return await self._interaction._get_trades_pair(False, address, trade_type, order_by, offset, limit)

    async def get_ohlcv(self, address: str, address_type: str, timeframe: str, dt_from: datetime, dt_to: datetime | None = None) -> GetOHLCVTokenPairResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[OHLCV - Token/Pair](https://docs.birdeye.so/reference/get_defi-ohlcv)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_ohlcv`][cyhole.birdeye.interaction.Birdeye._get_ohlcv].
        """
        return await self._interaction._get_ohlcv(False, address, address_type, timeframe, dt_from, dt_to)

    async def get_ohlcv_base_quote(self, base_address: str, quote_address: str, timeframe: str, dt_from: datetime, dt_to: datetime | None = None) -> GetOHLCVBaseQuoteResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[OHLCV - Base/Quote](https://docs.birdeye.so/reference/get_defi-ohlcv-base-quote)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_ohlcv_base_quote`][cyhole.birdeye.interaction.Birdeye._get_ohlcv_base_quote].
        """
        return await self._interaction._get_ohlcv_base_quote(False, base_address, quote_address, timeframe, dt_from, dt_to)

    async def get_wallet_supported_networks(self) -> GetWalletSupportedNetworksResponse:
        """
            Call the Birdeye's **PRIVATE** API endpoint **[Wallet - Supported Networks](https://docs.birdeye.so/reference/get_v1-wallet-list-supported-chain)** for asynchronous logic. 
            All the API endopint details are available on [`Birdeye._get_wallet_supported_networks`][cyhole.birdeye.interaction.Birdeye._get_wallet_supported_networks].
        """
        return await self._interaction._get_wallet_supported_networks(False)


from typing import Any
from datetime import datetime

from pydantic import BaseModel, Field, field_validator

# classes used on GET "Token List" endpoint
class GetTokenListInfo(BaseModel):
    name: str | None                = None
    symbol: str | None              = None
    address: str
    decimals: int
    liquidity: float
    volume_24h_usd: float           = Field(alias = "v24hUSD")
    market_cap: float               = Field(alias = "mc")
    volume_24h_change: float | None = Field(alias = "v24hChangePercent", default = None)
    last_trade_unix_time: float     = Field(alias = "lastTradeUnixTime")
    logo_uri: str  | None           = Field(alias = "logoURI", default = None)

class GetTokenListData(BaseModel):
    total: int
    update_time: datetime = Field(alias = "updateTime")
    update_unix_time: int = Field(alias = "updateUnixTime")
    tokens: list[GetTokenListInfo]

    @field_validator("update_time")
    def parse_update_time(cls, update_time_raw: str | datetime) -> datetime:
        if isinstance(update_time_raw, str):
            return datetime.strptime(update_time_raw, "%Y-%m-%dT%H:%M:%S")
        return update_time_raw

class GetTokenListResponse(BaseModel):
    """
        Model used to represent the **Token - List** endpoint from birdeye API.
    """
    data: GetTokenListData
    success: bool

# classes used on GET "Token - Security" endpoint
class GetTokenSecurityDataSolana(BaseModel):
    """
        Data specific only for Solana chain
    """
    creator_address: str | None = Field(alias = "creatorAddress", default = None)
    owner_address: str | None = Field(alias = "ownerAddress", default = None)
    creation_transaction: str | None = Field(alias = "creationTx", default = None)
    creation_time_unix: int | None = Field(alias = "creationTime", default = None)
    creation_slot: int | None = Field(alias = "creationSlot", default = None)
    mint_transaction: str | None = Field(alias = "mintTx", default = None)
    mint_time_unix: int | None = Field(alias = "mintTime", default = None)
    mint_slot: int | None = Field(alias = "mintSlot", default = None)
    mintable: bool | None = None
    renounce: bool | None = None
    creator_balance: float | None = Field(alias = "creatorBalance", default = None)
    creator_percentage: float | None = Field(alias = "creatorPercentage", default = None)
    owner_balance: float | None = Field(alias = "ownerBalance", default = None)
    owner_percentage: float | None = Field(alias = "ownerPercentage", default = None)
    metaplex_update_authority: str = Field(alias = "metaplexUpdateAuthority")
    metaplex_update_balance: float = Field(alias = "metaplexUpdateAuthorityBalance")
    metaplex_update_percentage: float = Field(alias = "metaplexUpdateAuthorityPercent")
    mutable_metadata: bool = Field(alias = "mutableMetadata")
    top_10_holder_balance: float = Field(alias = "top10HolderBalance")
    top_10_holder_percentage: float = Field(alias = "top10HolderPercent")
    top_10_user_balance: float = Field(alias = "top10UserBalance")
    top_10_user_percentage: float = Field(alias = "top10UserPercent")
    is_true_token: bool | None = Field(alias = "isTrueToken", default = None)
    total_supply: float = Field(alias = "totalSupply")
    pre_market_holder: list[str] = Field(alias = "preMarketHolder")
    lock_info: Any | None = Field(alias = "lockInfo", default = None)
    freezeable: bool | None = None
    freeze_authority: str | None = Field(alias = "freezeAuthority", default = None)
    transfer_fee_enable: bool | None = Field(alias = "transferFeeEnable", default = None)
    transfer_fee_data: Any | None = Field(alias = "transferFeeData", default = None)
    is_token_2022: bool = Field(alias = "isToken2022")
    non_transferable: bool | None = Field(alias = "nonTransferable", default = None)

class GetTokenSecurityResponse(BaseModel):
    """
        Model used to represent the **Token - Security** endpoint from birdeye API.
    """
    data: GetTokenSecurityDataSolana | dict[str, Any]
    success: bool

# classes used on GET "Token - Overview" endpoint
class GetTokenOverviewData(BaseModel):
    address: str
    decimals: int
    symbol: str
    name: str
    extensions: dict[str, str | None]
    logo_uri: str = Field(alias = "logoURI")
    liquidity: float
    price: float
    supply: float | None = None
    market_cap: float | None = Field(alias = "mc", default = None)
    history_30m_price: float = Field(alias = "history30mPrice")
    price_change_30m_percent: float = Field(alias = "priceChange30mPercent")
    history_1h_price: float = Field(alias = "history1hPrice")
    price_change_1h_percent: float = Field(alias = "priceChange1hPercent")
    history_2h_price: float = Field(alias = "history2hPrice")
    price_change_2h_percent: float = Field(alias = "priceChange2hPercent")
    history_4h_price: float = Field(alias = "history4hPrice")
    price_change_4h_percent: float = Field(alias = "priceChange4hPercent")
    history_6h_price: float = Field(alias = "history6hPrice")
    price_change_6h_percent: float = Field(alias = "priceChange6hPercent")
    history_8h_price: float = Field(alias = "history8hPrice")
    price_change_8h_percent: float = Field(alias = "priceChange8hPercent")
    history_12h_price: float = Field(alias = "history12hPrice")
    price_change_12h_percent: float = Field(alias = "priceChange12hPercent")
    history_24h_price: float = Field(alias = "history24hPrice")
    price_change_24h_percent: float = Field(alias = "priceChange24hPercent")
    unique_wallet_30m: int = Field(alias = "uniqueWallet30m")
    unique_wallet_history_30m: int = Field(alias = "uniqueWalletHistory30m")
    unique_wallet_30m_change_percent: float = Field(alias = "uniqueWallet30mChangePercent")
    unique_wallet_1h: int = Field(alias = "uniqueWallet1h")
    unique_wallet_history_1h: int = Field(alias = "uniqueWalletHistory1h")
    unique_wallet_1h_change_percent: float = Field(alias = "uniqueWallet1hChangePercent")
    unique_wallet_2h: int = Field(alias = "uniqueWallet2h")
    unique_wallet_history_2h: int = Field(alias = "uniqueWalletHistory2h")
    unique_wallet_2h_change_percent: float = Field(alias = "uniqueWallet2hChangePercent")
    unique_wallet_4h: int = Field(alias = "uniqueWallet4h")
    unique_wallet_history_4h: int = Field(alias = "uniqueWalletHistory4h")
    unique_wallet_4h_change_percent: float = Field(alias = "uniqueWallet4hChangePercent")
    unique_wallet_6h: int = Field(alias = "uniqueWallet6h")
    unique_wallet_history_6h: int = Field(alias = "uniqueWalletHistory6h")
    unique_wallet_6h_change_percent: float = Field(alias = "uniqueWallet6hChangePercent")
    unique_wallet_8h: int = Field(alias = "uniqueWallet8h")
    unique_wallet_history_8h: int = Field(alias = "uniqueWalletHistory8h")
    unique_wallet_8h_change_percent: float = Field(alias = "uniqueWallet8hChangePercent")
    unique_wallet_12h: int = Field(alias = "uniqueWallet12h")
    unique_wallet_history_12h: int = Field(alias = "uniqueWalletHistory12h")
    unique_wallet_12h_change_percent: float = Field(alias = "uniqueWallet12hChangePercent")
    unique_wallet_24h: int = Field(alias = "uniqueWallet24h")
    unique_wallet_history_24h: int = Field(alias = "uniqueWalletHistory24h")
    unique_wallet_24h_change_percent: float = Field(alias = "uniqueWallet24hChangePercent")
    last_trade_unix_time: int = Field(alias = "lastTradeUnixTime")
    last_trade_human_time: datetime = Field(alias = "lastTradeHumanTime")
    trade_30m: int = Field(alias = "trade30m")
    trade_history_30m: int = Field(alias = "tradeHistory30m")
    trade_30m_change_percent: float = Field(alias = "trade30mChangePercent")
    sell_30m: int = Field(alias = "sell30m")
    sell_history_30m: int = Field(alias = "sellHistory30m")
    sell_30m_change_percent: float = Field(alias = "sell30mChangePercent")
    buy_30m: int = Field(alias = "buy30m")
    buy_history_30m: int = Field(alias = "buyHistory30m")
    buy_30m_change_percent: float = Field(alias = "buy30mChangePercent")
    volume_30m: float = Field(alias = "v30m")
    volume_30m_usd: float = Field(alias = "v30mUSD")
    volume_history_30m: float = Field(alias = "vHistory30m")
    volume_history_30m_usd: float = Field(alias = "vHistory30mUSD")
    volume_30m_change_percent: float = Field(alias = "v30mChangePercent")
    volume_buy_30m: float = Field(alias = "vBuy30m")
    volume_buy_30m_usd: float = Field(alias = "vBuy30mUSD")
    volume_buy_history_30m: float = Field(alias = "vBuyHistory30m")
    volume_buy_history_30m_usd: float = Field(alias = "vBuyHistory30mUSD")
    volume_buy_30m_change_percent: float = Field(alias = "vBuy30mChangePercent")
    volume_sell_30m: float = Field(alias = "vSell30m")
    volume_sell_30m_usd: float = Field(alias = "vSell30mUSD")
    volume_sell_history_30m: float = Field(alias = "vSellHistory30m")
    volume_sell_history_30m_usd: float = Field(alias = "vSellHistory30mUSD")
    volume_sell_30m_change_percent: float = Field(alias = "vSell30mChangePercent")
    trade_1h: int = Field(alias = "trade1h")
    trade_history_1h: int = Field(alias = "tradeHistory1h")
    trade_1h_change_percent: float = Field(alias = "trade1hChangePercent")
    sell_1h: int = Field(alias = "sell1h")
    sell_history_1h: int = Field(alias = "sellHistory1h")
    sell_1h_change_percent: float = Field(alias = "sell1hChangePercent")
    buy_1h: int = Field(alias = "buy1h")
    buy_history_1h: int = Field(alias = "buyHistory1h")
    buy_1h_change_percent: float = Field(alias = "buy1hChangePercent")
    volume_1h: float = Field(alias = "v1h")
    volume_1h_usd: float = Field(alias = "v1hUSD")
    volume_history_1h: float = Field(alias = "vHistory1h")
    volume_history_1h_usd: float = Field(alias = "vHistory1hUSD")
    volume_1h_change_percent: float = Field(alias = "v1hChangePercent")
    volume_buy_1h: float = Field(alias = "vBuy1h")
    volume_buy_1h_usd: float = Field(alias = "vBuy1hUSD")
    volume_buy_history_1h: float = Field(alias = "vBuyHistory1h")
    volume_buy_history_1h_usd: float = Field(alias = "vBuyHistory1hUSD")
    volume_buy_1h_change_percent: float = Field(alias = "vBuy1hChangePercent")
    volume_sell_1h: float = Field(alias = "vSell1h")
    volume_sell_1h_usd: float = Field(alias = "vSell1hUSD")
    volume_sell_history_1h: float = Field(alias = "vSellHistory1h")
    volume_sell_history_1h_usd: float = Field(alias = "vSellHistory1hUSD")
    volume_sell_1h_change_percent: float = Field(alias = "vSell1hChangePercent")
    trade_2h: int = Field(alias = "trade2h")
    trade_history_2h: int = Field(alias = "tradeHistory2h")
    trade_2h_change_percent: float = Field(alias = "trade2hChangePercent")
    sell_2h: int = Field(alias = "sell2h")
    sell_history_2h: int = Field(alias = "sellHistory2h")
    sell_2h_change_percent: float = Field(alias = "sell2hChangePercent")
    buy_2h: int = Field(alias = "buy2h")
    buy_history_2h: int = Field(alias = "buyHistory2h")
    buy_2h_change_percent: float = Field(alias = "buy2hChangePercent")
    volume_2h: float = Field(alias = "v2h")
    volume_2h_usd: float = Field(alias = "v2hUSD")
    volume_history_2h: float = Field(alias = "vHistory2h")
    volume_history_2h_usd: float = Field(alias = "vHistory2hUSD")
    volume_2h_change_percent: float = Field(alias = "v2hChangePercent")
    volume_buy_2h: float = Field(alias = "vBuy2h")
    volume_buy_2h_usd: float = Field(alias = "vBuy2hUSD")
    volume_buy_history_2h: float = Field(alias = "vBuyHistory2h")
    volume_buy_history_2h_usd: float = Field(alias = "vBuyHistory2hUSD")
    volume_buy_2h_change_percent: float = Field(alias = "vBuy2hChangePercent")
    volume_sell_2h: float = Field(alias = "vSell2h")
    volume_sell_2h_usd: float = Field(alias = "vSell2hUSD")
    volume_sell_history_2h: float = Field(alias = "vSellHistory2h")
    volume_sell_history_2h_usd: float = Field(alias = "vSellHistory2hUSD")
    volume_sell_2h_change_percent: float = Field(alias = "vSell2hChangePercent")
    trade_4h: int = Field(alias = "trade4h")
    trade_history_4h: int = Field(alias = "tradeHistory4h")
    trade_4h_change_percent: float = Field(alias = "trade4hChangePercent")
    sell_4h: int = Field(alias = "sell4h")
    sell_history_4h: int = Field(alias = "sellHistory4h")
    sell_4h_change_percent: float = Field(alias = "sell4hChangePercent")
    buy_4h: int = Field(alias = "buy4h")
    buy_history_4h: int = Field(alias = "buyHistory4h")
    buy_4h_change_percent: float = Field(alias = "buy4hChangePercent")
    volume_4h: float = Field(alias = "v4h")
    volume_4h_usd: float = Field(alias = "v4hUSD")
    volume_history_4h: float = Field(alias = "vHistory4h")
    volume_history_4h_usd: float = Field(alias = "vHistory4hUSD")
    volume_4h_change_percent: float = Field(alias = "v4hChangePercent")
    volume_buy_4h: float = Field(alias = "vBuy4h")
    volume_buy_4h_usd: float = Field(alias = "vBuy4hUSD")
    volume_buy_history_4h: float = Field(alias = "vBuyHistory4h")
    volume_buy_history_4h_usd: float = Field(alias = "vBuyHistory4hUSD")
    volume_buy_4h_change_percent: float = Field(alias = "vBuy4hChangePercent")
    volume_sell_4h: float = Field(alias = "vSell4h")
    volume_sell_4h_usd: float = Field(alias = "vSell4hUSD")
    volume_sell_history_4h: float = Field(alias = "vSellHistory4h")
    volume_sell_history_4h_usd: float = Field(alias = "vSellHistory4hUSD")
    volume_sell_4h_change_percent: float = Field(alias = "vSell4hChangePercent")
    trade_6h: int = Field(alias = "trade6h")
    trade_history_6h: int = Field(alias = "tradeHistory6h")
    trade_6h_change_percent: float = Field(alias = "trade6hChangePercent")
    sell_6h: int = Field(alias = "sell6h")
    sell_history_6h: int = Field(alias = "sellHistory6h")
    sell_6h_change_percent: float = Field(alias = "sell6hChangePercent")
    buy_6h: int = Field(alias = "buy6h")
    buy_history_6h: int = Field(alias = "buyHistory6h")
    buy_6h_change_percent: float = Field(alias = "buy6hChangePercent")
    volume_6h: float = Field(alias = "v6h")
    volume_6h_usd: float = Field(alias = "v6hUSD")
    volume_history_6h: float = Field(alias = "vHistory6h")
    volume_history_6h_usd: float = Field(alias = "vHistory6hUSD")
    volume_6h_change_percent: float = Field(alias = "v6hChangePercent")
    volume_buy_6h: float = Field(alias = "vBuy6h")
    volume_buy_6h_usd: float = Field(alias = "vBuy6hUSD")
    volume_buy_history_6h: float = Field(alias = "vBuyHistory6h")
    volume_buy_history_6h_usd: float = Field(alias = "vBuyHistory6hUSD")
    volume_buy_6h_change_percent: float = Field(alias = "vBuy6hChangePercent")
    volume_sell_6h: float = Field(alias = "vSell6h")
    volume_sell_6h_usd: float = Field(alias = "vSell6hUSD")
    volume_sell_history_6h: float = Field(alias = "vSellHistory6h")
    volume_sell_history_6h_usd: float = Field(alias = "vSellHistory6hUSD")
    volume_sell_6h_change_percent: float = Field(alias = "vSell6hChangePercent")
    trade_8h: int = Field(alias = "trade8h")
    trade_history_8h: int = Field(alias = "tradeHistory8h")
    trade_8h_change_percent: float = Field(alias = "trade8hChangePercent")
    sell_8h: int = Field(alias = "sell8h")
    sell_history_8h: int = Field(alias = "sellHistory8h")
    sell_8h_change_percent: float = Field(alias = "sell8hChangePercent")
    buy_8h: int = Field(alias = "buy8h")
    buy_history_8h: int = Field(alias = "buyHistory8h")
    buy_8h_change_percent: float = Field(alias = "buy8hChangePercent")
    volume_8h: float = Field(alias = "v8h")
    volume_8h_usd: float = Field(alias = "v8hUSD")
    volume_history_8h: float = Field(alias = "vHistory8h")
    volume_history_8h_usd: float = Field(alias = "vHistory8hUSD")
    volume_8h_change_percent: float = Field(alias = "v8hChangePercent")
    volume_buy_8h: float = Field(alias = "vBuy8h")
    volume_buy_8h_usd: float = Field(alias = "vBuy8hUSD")
    volume_buy_history_8h: float = Field(alias = "vBuyHistory8h")
    volume_buy_history_8h_usd: float = Field(alias = "vBuyHistory8hUSD")
    volume_buy_8h_change_percent: float = Field(alias = "vBuy8hChangePercent")
    volume_sell_8h: float = Field(alias = "vSell8h")
    volume_sell_8h_usd: float = Field(alias = "vSell8hUSD")
    volume_sell_history_8h: float = Field(alias = "vSellHistory8h")
    volume_sell_history_8h_usd: float = Field(alias = "vSellHistory8hUSD")
    volume_sell_8h_change_percent: float = Field(alias = "vSell8hChangePercent")
    trade_12h: int = Field(alias = "trade12h")
    trade_history_12h: int = Field(alias = "tradeHistory12h")
    trade_12h_change_percent: float = Field(alias = "trade12hChangePercent")
    sell_12h: int = Field(alias = "sell12h")
    sell_history_12h: int = Field(alias = "sellHistory12h")
    sell_12h_change_percent: float = Field(alias = "sell12hChangePercent")
    buy_12h: int = Field(alias = "buy12h")
    buy_history_12h: int = Field(alias = "buyHistory12h")
    buy_12h_change_percent: float = Field(alias = "buy12hChangePercent")
    volume_12h: float = Field(alias = "v12h")
    volume_12h_usd: float = Field(alias = "v12hUSD")
    volume_history_12h: float = Field(alias = "vHistory12h")
    volume_history_12h_usd: float = Field(alias = "vHistory12hUSD")
    volume_12h_change_percent: float = Field(alias = "v12hChangePercent")
    volume_buy_12h: float = Field(alias = "vBuy12h")
    volume_buy_12h_usd: float = Field(alias = "vBuy12hUSD")
    volume_buy_history_12h: float = Field(alias = "vBuyHistory12h")
    volume_buy_history_12h_usd: float = Field(alias = "vBuyHistory12hUSD")
    volume_buy_12h_change_percent: float = Field(alias = "vBuy12hChangePercent")
    volume_sell_12h: float = Field(alias = "vSell12h")
    volume_sell_12h_usd: float = Field(alias = "vSell12hUSD")
    volume_sell_history_12h: float = Field(alias = "vSellHistory12h")
    volume_sell_history_12h_usd: float = Field(alias = "vSellHistory12hUSD")
    volume_sell_12h_change_percent: float = Field(alias = "vSell12hChangePercent")
    trade_24h: int = Field(alias = "trade24h")
    trade_history_24h: int = Field(alias = "tradeHistory24h")
    trade_24h_change_percent: float = Field(alias = "trade24hChangePercent")
    sell_24h: int = Field(alias = "sell24h")
    sell_history_24h: int = Field(alias = "sellHistory24h")
    sell_24h_change_percent: float = Field(alias = "sell24hChangePercent")
    buy_24h: int = Field(alias = "buy24h")
    buy_history_24h: int = Field(alias = "buyHistory24h")
    buy_24h_change_percent: float = Field(alias = "buy24hChangePercent")
    volume_24h: float = Field(alias = "v24h")
    volume_24h_usd: float = Field(alias = "v24hUSD")
    volume_history_24h: float = Field(alias = "vHistory24h")
    volume_history_24h_usd: float = Field(alias = "vHistory24hUSD")
    volume_24h_change_percent: float = Field(alias = "v24hChangePercent")
    volume_buy_24h: float = Field(alias = "vBuy24h")
    volume_buy_24h_usd: float = Field(alias = "vBuy24hUSD")
    volume_buy_history_24h: float = Field(alias = "vBuyHistory24h")
    volume_buy_history_24h_usd: float = Field(alias = "vBuyHistory24hUSD")
    volume_buy_24h_change_percent: float = Field(alias = "vBuy24hChangePercent")
    volume_sell_24h: float = Field(alias = "vSell24h")
    volume_sell_24h_usd: float = Field(alias = "vSell24hUSD")
    volume_sell_history_24h: float = Field(alias = "vSellHistory24h")
    volume_sell_history_24h_usd: float = Field(alias = "vSellHistory24hUSD")
    volume_sell_24h_change_percent: float = Field(alias = "vSell24hChangePercent")
    watch: int = Field(alias = "watch")
    view_30m: int = Field(alias = "view30m")
    view_history_30m: int = Field(alias = "viewHistory30m")
    view_30m_change_percent: float = Field(alias = "view30mChangePercent")
    view_1h: int = Field(alias = "view1h")
    view_history_1h: int = Field(alias = "viewHistory1h")
    view_1h_change_percent: float = Field(alias = "view1hChangePercent")
    view_2h: int = Field(alias = "view2h")
    view_history_2h: int = Field(alias = "viewHistory2h")
    view_2h_change_percent: float = Field(alias = "view2hChangePercent")
    view_4h: int = Field(alias = "view4h")
    view_history_4h: int = Field(alias = "viewHistory4h")
    view_4h_change_percent: float = Field(alias = "view4hChangePercent")
    view_6h: int = Field(alias = "view6h")
    view_history_6h: int = Field(alias = "viewHistory6h")
    view_6h_change_percent: float = Field(alias = "view6hChangePercent")
    view_8h: int = Field(alias = "view8h")
    view_history_8h: int = Field(alias = "viewHistory8h")
    view_8h_change_percent: float = Field(alias = "view8hChangePercent")
    view_12h: int = Field(alias = "view12h")
    view_history_12h: int = Field(alias = "viewHistory12h")
    view_12h_change_percent: float = Field(alias = "view12hChangePercent")
    view_24h: int = Field(alias = "view24h")
    view_history_24h: int = Field(alias = "viewHistory24h")
    view_24h_change_percent: float = Field(alias = "view24hChangePercent")
    unique_view_30m: int = Field(alias = "uniqueView30m")
    unique_view_history_30m: int = Field(alias = "uniqueViewHistory30m")
    unique_view_30m_change_percent: float = Field(alias = "uniqueView30mChangePercent")
    unique_view_1h: int = Field(alias = "uniqueView1h")
    unique_view_history_1h: int = Field(alias = "uniqueViewHistory1h")
    unique_view_1h_change_percent: float = Field(alias = "uniqueView1hChangePercent")
    unique_view_2h: int = Field(alias = "uniqueView2h")
    unique_view_history_2h: int = Field(alias = "uniqueViewHistory2h")
    unique_view_2h_change_percent: float = Field(alias = "uniqueView2hChangePercent")
    unique_view_4h: int = Field(alias = "uniqueView4h")
    unique_view_history_4h: int = Field(alias = "uniqueViewHistory4h")
    unique_view_4h_change_percent: float = Field(alias = "uniqueView4hChangePercent")
    unique_view_6h: int = Field(alias = "uniqueView6h")
    unique_view_history_6h: int = Field(alias = "uniqueViewHistory6h")
    unique_view_6h_change_percent: float = Field(alias = "uniqueView6hChangePercent")
    unique_view_8h: int = Field(alias = "uniqueView8h")
    unique_view_history_8h: int = Field(alias = "uniqueViewHistory8h")
    unique_view_8h_change_percent: float = Field(alias = "uniqueView8hChangePercent")
    unique_view_12h: int = Field(alias = "uniqueView12h")
    unique_view_history_12h: int = Field(alias = "uniqueViewHistory12h")
    unique_view_12h_change_percent: float = Field(alias = "uniqueView12hChangePercent")
    unique_view_24h: int = Field(alias = "uniqueView24h")
    unique_view_history_24h: int = Field(alias = "uniqueViewHistory24h")
    unique_view_24h_change_percent: float = Field(alias = "uniqueView24hChangePercent")

class GetTokenOverviewResponse(BaseModel):
    """
        Model used to represent the **Token - Overview** endpoint from birdeye API.
    """
    data: GetTokenOverviewData
    success: bool

# classes used on GET "Token - Creation Token Info" endpoint
class GetTokenCreationInfoData(BaseModel):
    transaction_hash: str = Field(alias = "txHash")
    slot: int
    token_address: str = Field(alias = "tokenAddress")
    decimals: int
    owner: str

class GetTokenCreationInfoResponse(BaseModel):
    """
        Model used to represent the **Token - Creation Token Info** endpoint from birdeye API.
    """
    data: GetTokenCreationInfoData
    success: bool


# classes used on GET "Price" endpoint
class GetPriceData(BaseModel):
    value: float
    liquidity: float | None = None
    update_human_time: datetime = Field(alias = "updateHumanTime")
    update_unix_time: int = Field(alias = "updateUnixTime")

    @field_validator("update_human_time")
    def parse_update_human_time(cls, update_human_time_raw: str | datetime) -> datetime:
        if isinstance(update_human_time_raw, str):
            return datetime.strptime(update_human_time_raw, "%Y-%m-%dT%H:%M:%S")
        return update_human_time_raw

class GetPriceResponse(BaseModel):
    """
        Model used to represent the **Price** endpoint from birdeye API.
    """
    data: GetPriceData
    success: bool

# classes used on GET "Price - Multiple" endpoint
class GetPriceMultipleData(GetPriceData):
    price_change_24h: float = Field(alias = "priceChange24h")

class GetPriceMultipleResponse(BaseModel):
    """
        Model used to represent the **Price - Multiple** endpoint from birdeye API.
    """
    data: dict[str, GetPriceMultipleData]
    success: bool

# classes used on GET "Price - Historical" endpoint
class GetPriceHistoricalMeasure(BaseModel):
    unix_time: int = Field(alias = "unixTime")
    value: float

class GetPriceHistoricalData(BaseModel):
    items: list[GetPriceHistoricalMeasure]

class GetPriceHistoricalResponse(BaseModel):
    """
        Model used to represent the **Price - Historical** endpoint from birdeye API.
    """
    data: GetPriceHistoricalData
    success: bool

# classes used on GET "Trades - Token" endpoint
class GetTradesTokenTradeToken(BaseModel):
    symbol: str
    decimals: int
    address: str
    amount: int
    type: str
    type_swap: str = Field(alias = "typeSwap")
    ui_amount: float = Field(alias = "uiAmount")
    price: float | None = None
    nearest_price: float = Field(alias = "nearestPrice")
    change_amount: float = Field(alias = "changeAmount")
    ui_change_amount: float = Field(alias = "uiChangeAmount")
    icon: str | None = None

class GetTradesTokenTrade(BaseModel):
    volume: float
    volume_usd: float = Field(alias = "volumeUSD")
    trade_hash: str = Field(alias = "txHash")
    slot: int
    source: str
    block_unix_time: int = Field(alias = "blockUnixTime")
    block_human_time: datetime = Field(alias = "blockHumanTime")
    trade_type: str = Field(alias = "txType")
    address: str
    owner: str
    trade_status: str = Field(alias = "txStatus")
    outliers: bool
    nearest_price_base_coin: float = Field(alias = "nearestPriceBaseCoin")
    nearest_price_quote_coin: float = Field(alias = "nearestPriceQuoteCoin")
    ins_index: int = Field(alias = "insIndex")
    inner_ins_index: int | None = Field(alias = "innerInsIndex", default = None)
    be_bevenue: str | None = Field(alias = "beRevenue", default = None)
    event_type: str = Field(alias = "eventType")
    side: str
    price_pair: float = Field(alias = "pricePair")
    alias: str | None = None
    platform: str
    toke_price: float = Field(alias = "tokenPrice")
    proce_mark: bool = Field(alias = "priceMark")
    network: str
    trade_from: GetTradesTokenTradeToken = Field(alias = "from")
    trade_to: GetTradesTokenTradeToken = Field(alias = "to")

    @field_validator("block_human_time")
    def parse_update_human_time(cls, dt_raw: str | datetime) -> datetime:
        if isinstance(dt_raw, str):
            return datetime.strptime(dt_raw, "%Y-%m-%dT%H:%M:%S")
        return dt_raw

class GetTradesTokenData(BaseModel):
    items: list[GetTradesTokenTrade]
    has_next: bool = Field(alias = "hasNext")

class GetTradesTokenResponse(BaseModel):
    """
        Model used to represent the **Trades - Token** endpoint from birdeye API.
    """
    data: GetTradesTokenData
    success: bool

# classes used on GET "Trades - Pair" endpoint
class GetTradesPairTradeToken(BaseModel):
    symbol: str
    decimals: int
    address: str
    amount: int
    type: str
    type_swap: str = Field(alias = "typeSwap")
    ui_amount: float = Field(alias = "uiAmount")
    price: float | None = None
    nearest_price: float = Field(alias = "nearestPrice")
    change_amount: float = Field(alias = "changeAmount")
    ui_change_amount: float = Field(alias = "uiChangeAmount")

class GetTradesPairTrade(BaseModel):
    trade_hash: str = Field(alias = "txHash")
    source: str
    block_unix_time: int = Field(alias = "blockUnixTime")
    address: str
    owner: str
    trade_from: GetTradesPairTradeToken = Field(alias = "from")
    trade_to: GetTradesPairTradeToken = Field(alias = "to")

class GetTradesPairData(BaseModel):
    items: list[GetTradesPairTrade]
    has_next: bool = Field(alias = "hasNext")

class GetTradesPairResponse(BaseModel):
    """
        Model used to represent the **Trades - Pair** endpoint from birdeye API.
    """
    data: GetTradesPairData
    success: bool

# classes used on GET "OHLCV - Token/Pair" endpoint
class GetOHLCVInterval(BaseModel):
    close: float = Field(alias = "c")
    high: float  = Field(alias = "h")
    low: float  = Field(alias = "l")
    open: float  = Field(alias = "o")
    type: str
    unix_time: int  = Field(alias = "unixTime")

class GetOHLCVTokenPairInterval(GetOHLCVInterval):
    address: str
    volume: float  = Field(alias = "v")

class GetOHLCVTokenPairData(BaseModel):
    items: list[GetOHLCVTokenPairInterval]

class GetOHLCVTokenPairResponse(BaseModel):
    """
        Model used to represent the **OHLCV - Token/Pair** endpoint from birdeye API.
    """
    data: GetOHLCVTokenPairData
    success: bool

# classes used on GET "OHLCV - Base/Quote" endpoint
class GetOHLCVBaseQuoteInterval(GetOHLCVInterval):
    base_address: str = Field(alias = "baseAddress")
    quote_address: str = Field(alias = "quoteAddress")
    base_volume: float = Field(alias = "vBase")
    quote_volume: float = Field(alias = "vQuote")

class GetOHLCVBaseQuoteData(BaseModel):
    items: list[GetOHLCVBaseQuoteInterval]

class GetOHLCVBaseQuoteResponse(BaseModel):
    """
        Model used to represent the **OHLCV - Token/Pair** endpoint from birdeye API.
    """
    data: GetOHLCVBaseQuoteData
    success: bool

# classes used on GET "Wallet - Supported Networks" endpoint
class GetWalletSupportedNetworksResponse(BaseModel):
    """
        Model used to represent the **Wallet - Supported Networks** endpoint from birdeye API.
    """
    data: list[str]
    success: bool


from typing import Any

from ..core.client import APIClient, AsyncAPIClient

class Interaction:
    """
        This class is the building block for a new integration of an external API in `cyhole` library, 
        and every new external API **must** be identified by a class that inherits from this class.

        Inside an `Interaction` there are two attributes identifying the clients responsable to 
        perform the API calls:

        - `client`: object used for **synchronous** logic.
        - `async_client`: object used for **asynchronous** logic.

        During the creation of the object is possible to specify some global configurations.

        Parameters:
            headers: headers used globally in all API requests.
    """
    def __init__(self, headers: Any | None = None) -> None:
        self.headers = headers

        # clients
        self.client = APIClient(self, headers = self.headers)
        self.async_client = AsyncAPIClient(self, headers = self.headers)

        return


"""
    This file is used to load and manage all the tests' configurations.

    The 'cyhole' library interacts with many external API that requires 
    an API Key in order to extract data. Since the usage of these API Keys 
    is subjected to costs, a proper configuration for the tests was required.

    The configuration is managed by a 'test.ini' file loaded during the 
    execution of the tests. By using this file will be possible to:
        - manage API Keys externally
        - enable/disable mock responses

    Check the default 'test.ini' file for a complete list of available configurations.

    Mock Responses
    If the mock response functionality is enabled, then the tests try to load a JSON file 
    with the response stored from a previous extration from the API.
    Moreover, every time the tests are executed with the mock response functionality 
    disabled, the process automatically stores the response in a JSON file into the 
    proper resources folder.

"""
import json
import configparser
from pathlib import Path
from requests import Response
from typing import TypeVar, Type
from pydantic import BaseModel

ResponseModel = TypeVar('ResponseModel', bound = BaseModel)

class JupiterConfiguration(BaseModel):
    """
        Model in charge to manage the Jupiter APIs.
    """
    mock_response: bool = True
    mock_folder: str = "jupiter"

class BirdeyeConfiguration(BaseModel):
    """
        Model in charge to manage the birdeye APIs.
    """
    mock_response_public: bool = False
    mock_response_private: bool = True
    mock_folder: str = "birdeye"
    api_key: str = ""

class TestConfiguration(BaseModel):
    """
        Model in charge to manage the tests' configuration.
    """
    mock_response: bool = True
    mock_folder: str = "tests/resources/mock"
    mock_file_overwrite: bool = False
    birdeye: BirdeyeConfiguration = BirdeyeConfiguration()
    jupiter: JupiterConfiguration = JupiterConfiguration()

def load_config(path: str = "tests", file: str = "test.ini") -> TestConfiguration:
    """
        This function is used to load the test configuration file.

        It is strongly suggested to leave the default 'path' and 'file' 
        name for the test configuration file. However, if a dedicated file 
        is used, then is possible to load another file by providing the 
        necesary information in the input variables.

        Parameters:
            path: path of the configuration file.
            file: name of the configuration file (with extension).

        Returns:
            pydantic model of the test's configuration.
    """
    config_path_file = f"{path}/{file}"
    config = configparser.ConfigParser()

    # load config
    config.read(config_path_file)
    test_config = TestConfiguration()
    
    # global
    test_config.mock_response = config.getboolean("global", "mock_response", fallback = test_config.mock_response)
    test_config.mock_folder = config.get("global", "mock_folder", fallback = test_config.mock_folder)
    test_config.mock_file_overwrite = config.getboolean("global", "mock_file_overwrite", fallback = test_config.mock_file_overwrite)

    # birdeye
    test_config.birdeye.mock_response_public = config.getboolean("birdeye", "mock_response_public", fallback = test_config.birdeye.mock_response_public)
    test_config.birdeye.mock_response_private = config.getboolean("birdeye", "mock_response_private", fallback = test_config.birdeye.mock_response_private)
    test_config.birdeye.mock_folder = config.get("birdeye", "mock_folder", fallback = test_config.birdeye.mock_folder)
    test_config.birdeye.api_key = config.get("birdeye", "api_key", fallback = test_config.birdeye.api_key)

    # jupiter
    test_config.jupiter.mock_response = config.getboolean("jupiter", "mock_response", fallback = test_config.jupiter.mock_response)
    test_config.jupiter.mock_folder = config.get("jupiter", "mock_folder", fallback = test_config.jupiter.mock_folder)

    return test_config

class MockerManager:
    """
        Class used to manage the mock responses.
    """

    def __init__(self, mock_path: Path) -> None:
        # set/create resources folder
        self.mock_path = mock_path
        mock_path.mkdir(parents = True, exist_ok = True)
    
    def load_mock_model(self, file_name: str, response_model: Type[ResponseModel]) -> ResponseModel:
        """
            Use this function to load mock response model from a file.

            Parameters:
                file_name: file name containing the response to load.
                    By assumption, the file must be a JSON file and the file name should not contain the extension.
                response_model: objects that inherit from `pydantic.BaseModel` and describe the response object element.
        """
        mock_file = f"{file_name}.json"
        mock_path_file = self.mock_path / mock_file

        if not mock_path_file.exists():
            raise Exception(f"mock file '{mock_file}' does not exist in '{self.mock_path}'.")

        with open(mock_path_file, "r") as file:
            data = json.loads(file.read())
            mock_response = response_model(**data)

        return mock_response

    def store_mock_model(self, file_name: str, response: BaseModel) -> None:
        """
            Use this function to store the mock response model into a file.

            Parameters:
                file_name: file name containing the response to load.
                    By assumption, the file must be a JSON file and the file name should not contain the extension.
                response: objects that inherit from `pydantic.BaseModel` and describe the response object element.
        """
        mock_file = f"{file_name}.json"
        mock_path_file = self.mock_path / mock_file

        with open(mock_path_file, "w") as file:
            file.write(response.model_dump_json(indent = 4, by_alias = True))

        return

    def load_mock_response(self, file_name: str, response_model: Type[ResponseModel]) -> Response:
        """
            Use this function to load mock response model from a file and return a dummy `Response` object.

            Parameters:
                file_name: file name containing the response to load.
                    By assumption, the file must be a JSON file and the file name should not contain the extension.
                response_model: objects that inherit from `pydantic.BaseModel` and describe the response object element.
        """
        mock_model = self.load_mock_model(file_name, response_model)

        # create dummy response
        mock_response = Response()
        mock_response._content = mock_model.model_dump_json(by_alias = True, exclude_none = True).encode()
        mock_response.status_code = 200
        mock_response.encoding = "utf-8"

        return mock_response


from ..core.exception import CyholeException

class JupiterException(CyholeException):
    """General Exception for Jupiter API."""
    pass

class JupiterNoRouteFoundError(JupiterException):
    pass

class JupiterInvalidRequest(JupiterException):
    pass


from ..core.param import CyholeParam

class JupiterSwapMode(CyholeParam):
    """
        Enum class to get the supported Jupiter's swap modes.
    """
    EXACT_IN = "ExactIn"
    EXACT_OUT = "ExactOut"

class JupiterSwapDex(CyholeParam):
    """
        Enum class to get the supported Jupiter's swap DEXes.
    """
    ALDRIN = "Aldrin"
    ALDRIN_V2 = "Aldrin V2"
    BONKSWAP = "Bonkswap"
    CLONE = "Clone Protocol"
    CREMA = "Crema"
    CROPPER = "Cropper"
    CROPPER_LEGACY = "Cropper Legacy"
    DEXLAB = "Dexlab"
    FLUX = "FluxBeam"
    GOOSEFX = "GooseFX"
    GUACSWAP = "Guacswap"
    HELIUM = "Helium Network"
    INVARIANT = "Invariant"
    LIFINITY_V1 = "Lifinity V1"
    LIFINITY_V2 = "Lifinity V2"
    MARINADE = "Marinade"
    MERCURIAL = "Mercurial"
    METEORA = "Meteora"
    METEORA_DLMM = "Meteora DLMM"
    OASIS = "Oasis"
    OPENBOOK = "Openbook"
    OPENBOOK_V2 = "OpenBook V2"
    ORCA_V1 = "Orca V1"
    ORCA_V2 = "Orca V2"
    PENGUIN = "Penguin"
    PERPS = "Perps"
    PHOENIX = "Phoenix"
    RAYDIUM = "Raydium"
    RAYDIUM_CLMM = "Raydium CLMM"
    RAYDIUM_CP = "Raydium CP"
    SABER = "Saber"
    SABER_DECIMALS = "Saber (Decimals)"
    SANCTUM = "Sanctum"
    SANCTUM_INFINITY = "Sanctum Infinity"
    SAROS = "Saros"
    STEP = "StepN"
    TOKEN_SWAP = "Token Swap"
    WHIRLPOOL = "Whirlpool"

class JupiterTokenListType(CyholeParam):
    """
        Enum class to get the supported Jupiter's token lists.
    """
    STRICT = "strict"
    """Extract only the tokens with tags `old-registry`, `community`, or `wormhole` verified"""
    ALL = "all"
    """Extract **all** non banned tokens."""

class JupiterLimitOrderState(CyholeParam):
    """
        Enum class to get the supported Jupiter's limit order states.
    """
    OPEN = "Open"
    """The limit order is still open."""
    COMPLETED = "Completed"
    """The limit order is completed."""
    CANCELLED = "Cancelled"
    """The limit order was cancelled."""


SOL = "So11111111111111111111111111111111111111112"
USDC = "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v"
USDT = "Es9vMFrzaCERmJfrF4H2FYD4KCoNkY11McCe8BenwNYB"
JUP = "JUPyiwrYJFskUPiHa7hkeR8VUtAeFoSYbKedZNsDvCN"
BONK = "DezXAZ8z7PnrnRJjz3wXBoRgixCa6xjnB7YaB1pPB263"
WIF = "EKpQGSJtjMFqKZ9KQanSqYXRcF8fBopzLHYxdM65zcjm"


from datetime import datetime, timedelta
from pathlib import Path

import pytest
from pytest_mock import MockerFixture

from cyhole.birdeye import Birdeye
from cyhole.birdeye.param import BirdeyeAddressType, BirdeyeTimeFrame, BirdeyeChain
from cyhole.birdeye.schema import (
    GetTokenListResponse,
    GetTokenCreationInfoResponse,
    GetTokenSecurityResponse, GetTokenSecurityDataSolana,
    GetTokenOverviewResponse,
    GetPriceResponse,
    GetPriceMultipleResponse,
    GetPriceHistoricalResponse,
    GetTradesTokenResponse,
    GetTradesPairResponse,
    GetOHLCVTokenPairResponse,
    GetOHLCVBaseQuoteResponse,
    GetWalletSupportedNetworksResponse
)
from cyhole.birdeye.exception import BirdeyeAuthorisationError, BirdeyeTimeRangeError
from cyhole.core.exception import MissingAPIKeyError
from cyhole.core.address.solana import SOL, USDC
from cyhole.core.address.ethereum import WETH

# load test config
from .config import load_config, MockerManager
config = load_config()

# constant address
TOM_SOL = "842NwDnKYcfMRWAYqsD3hoTWXKKMi28gVABtmaupFcnS"
JRK = "JRKXwVpdyQbF3A4pvQvKYj22syubbEwfwUiobDzSPtJ"

# create resources folder
mock_path = Path(config.mock_folder) / config.birdeye.mock_folder
mock_path.mkdir(parents = True, exist_ok = True)

class TestBirdeyePublic:
    """
        Class grouping all unit test associate to **PUBLIC** endpoints.
    """
    birdeye = Birdeye(api_key = config.birdeye.api_key)
    mocker = MockerManager(mock_path)

    def test_missing_api_key(self) -> None:
        """
            Unit Test to correcty identify a missing/wrong API Key.
        """
        with pytest.raises(MissingAPIKeyError):
            Birdeye()

    def test_get_token_list_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - List" 
            for synchronous logic.

            Mock Response File: get_token_list.json
        """
    
        # load mock response
        mock_file_name = "get_token_list"
        if config.mock_response or config.birdeye.mock_response_public:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenListResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)

        # execute request
        response = self.birdeye.client.get_token_list(limit = 1)

        # actual test
        assert isinstance(response, GetTokenListResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_public:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_list_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - List" 
            for asynchronous logic.

            Mock Response File: get_token_list.json
        """
    
        # load mock response
        mock_file_name = "get_token_list"
        if config.mock_response or config.birdeye.mock_response_public:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenListResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)

        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_token_list(limit = 1)

        # actual test
        assert isinstance(response, GetTokenListResponse)

    def test_get_price_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for synchronous logic.

            Mock Response File: get_price.json
        """

        # load mock response
        mock_file_name = "get_price"
        if config.mock_response or config.birdeye.mock_response_public:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_price(address = SOL)

        # actual test
        assert isinstance(response, GetPriceResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_public:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price" 
            for asynchronous logic.

            Mock Response File: get_price.json
        """

        # load mock response
        mock_file_name = "get_price"
        if config.mock_response or config.birdeye.mock_response_public:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_price(address = SOL)

        # actual test
        assert isinstance(response, GetPriceResponse)

    def test_get_price_historical_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price - Historical" 
            for synchronous logic.

            Mock Response File: get_price_historical.json
        """

        # load mock response
        mock_file_name = "get_price_historical"
        if config.mock_response or config.birdeye.mock_response_public:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceHistoricalResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_price_historical(
            address = SOL,
            address_type = BirdeyeAddressType.TOKEN.value,
            timeframe = BirdeyeTimeFrame.MIN15.value,
            dt_from = datetime.now() - timedelta(hours = 1)
        )

        # actual test
        assert isinstance(response, GetPriceHistoricalResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_public:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_historical_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price - Historical" 
            for asynchronous logic.

            Mock Response File: get_price_historical.json
        """

        # load mock response
        mock_file_name = "get_price_historical"
        if config.mock_response or config.birdeye.mock_response_public:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceHistoricalResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_price_historical(
                address = SOL,
                address_type = BirdeyeAddressType.TOKEN.value,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() - timedelta(hours = 1)
            )

        # actual test
        assert isinstance(response, GetPriceHistoricalResponse)

    def test_get_price_historical_incorrect_input_dates_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the incorrect dates inputs (dt_from > dt_to) 
            for synchronous logic.
        """

        with pytest.raises(BirdeyeTimeRangeError):
            # execute request
            self.birdeye.client.get_price_historical(
                address = SOL,
                address_type = BirdeyeAddressType.TOKEN.value,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() + timedelta(hours = 1)
            )

    @pytest.mark.asyncio
    async def test_get_price_historical_incorrect_input_dates_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the incorrect dates inputs (dt_from > dt_to) 
            for asynchronous logic.
        """

        with pytest.raises(BirdeyeTimeRangeError):
            # execute request
            async with self.birdeye.async_client as client:
                await client.get_price_historical(
                    address = SOL,
                    address_type = BirdeyeAddressType.TOKEN.value,
                    timeframe = BirdeyeTimeFrame.MIN15.value,
                    dt_from = datetime.now() + timedelta(hours = 1)
                )

class TestBirdeyePrivate:
    """
        Class grouping all unit test associate to PRIVATE endpoints
    """
    birdeye = Birdeye(api_key = config.birdeye.api_key)
    mocker = MockerManager(mock_path)

    def test_not_authorised_api_sync(self) -> None:
        """
            Unit Test to correcty identify a not Authorised API Key 
            for synchronous logic.
        """
        birdeye = Birdeye(api_key = "xxx-xxx-xxx")
        with pytest.raises(BirdeyeAuthorisationError):
            birdeye.client.get_token_creation_info(address = SOL)

    @pytest.mark.asyncio
    async def test_not_authorised_api_async(self) -> None:
        """
            Unit Test to correcty identify a not Authorised API Key 
            for asynchronous logic.
        """
        birdeye = Birdeye(api_key = "xxx-xxx-xxx")
        with pytest.raises(BirdeyeAuthorisationError):
            async with birdeye.async_client as client:
                await client.get_token_creation_info(address = SOL)

    def test_get_price_multiple_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price - Multiple" 
            for synchronous logic.

            Mock Response File: get_price_multiple.json
        """

        # load mock response
        mock_file_name = "get_price_multiple"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceMultipleResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        tokens_ca = [SOL, USDC]
        response = self.birdeye.client.get_price_multiple(list_address = tokens_ca)

        # actual test
        assert isinstance(response, GetPriceMultipleResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_price_multiple_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Price - Multiple" 
            for asynchronous logic.

            Mock Response File: get_price_multiple.json
        """

        # load mock response
        mock_file_name = "get_price_multiple"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetPriceMultipleResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        tokens_ca = [SOL, USDC]
        async with self.birdeye.async_client as client:
            response = await client.get_price_multiple(list_address = tokens_ca)

        # actual test
        assert isinstance(response, GetPriceMultipleResponse)

    def test_get_token_creation_info_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Creation Token Info" 
            for synchronous logic.

            Mock Response File: get_token_creation_info.json
        """

        # load mock response
        mock_file_name = "get_token_creation_info"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenCreationInfoResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_token_creation_info(address = SOL)

        # actual test
        assert isinstance(response, GetTokenCreationInfoResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_creation_info_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Creation Token Info" 
            for asynchronous logic.

            Mock Response File: get_token_creation_info.json
        """

        # load mock response
        mock_file_name = "get_token_creation_info"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenCreationInfoResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_token_creation_info(address = SOL)

        # actual test
        assert isinstance(response, GetTokenCreationInfoResponse)

    def test_get_token_security_solana_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Security" 
            specifically for Solana chain for synchronous logic.

            Mock Response File: get_token_security_solana.json
        """

        # load mock response
        mock_file_name = "get_token_security_solana"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenSecurityResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_token_security(JRK)

        # actual test
        assert isinstance(response, GetTokenSecurityResponse)
        assert isinstance(response.data, GetTokenSecurityDataSolana)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_security_solana_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Security" 
            specifically for Solana chain for asynchronous logic.

            Mock Response File: get_token_security_solana.json
        """

        # load mock response
        mock_file_name = "get_token_security_solana"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenSecurityResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_token_security(JRK)

        # actual test
        assert isinstance(response, GetTokenSecurityResponse)
        assert isinstance(response.data, GetTokenSecurityDataSolana)

    def test_get_token_security_other_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Security" 
            specifically for other chains for synchronous logic.

            Mock Response File: get_token_security_other.json
        """
        birdeye = Birdeye(api_key = config.birdeye.api_key, chain = BirdeyeChain.ETHEREUM.value)

        # load mock response
        mock_file_name = "get_token_security_other"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenSecurityResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = birdeye.client.get_token_security(address = JRK)

        # actual test
        assert isinstance(response, GetTokenSecurityResponse)
        assert isinstance(response.data, dict)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_security_other_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Security" 
            specifically for other chains for asynchronous logic.

            Mock Response File: get_token_security_other.json
        """
        birdeye = Birdeye(api_key = config.birdeye.api_key, chain = BirdeyeChain.ETHEREUM.value)

        # load mock response
        mock_file_name = "get_token_security_other"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenSecurityResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with birdeye.async_client as client:
            response = await client.get_token_security(address = JRK)

        # actual test
        assert isinstance(response, GetTokenSecurityResponse)
        assert isinstance(response.data, dict)

    def test_get_token_overview_solana_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Overview" 
            specifically for Solana chain for synchronous logic.

            Mock Response File: get_token_overview_solana.json
        """
        token_address = SOL

        # load mock response
        mock_file_name = "get_token_overview_solana"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenOverviewResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_token_overview(token_address)

        # actual test
        assert isinstance(response, GetTokenOverviewResponse)
        assert response.data.address == token_address

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_overview_solana_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Overview" 
            specifically for Solana chain for asynchronous logic.

            Mock Response File: get_token_overview_solana.json
        """
        token_address = SOL

        # load mock response
        mock_file_name = "get_token_overview_solana"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenOverviewResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_token_overview(token_address)

        # actual test
        assert isinstance(response, GetTokenOverviewResponse)
        assert response.data.address == token_address

    def test_get_token_overview_ethereum_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Overview" 
            specifically for Ethereum chain for synchronous logic.

            Mock Response File: get_token_overview_ethereum.json
        """
        birdeye = Birdeye(api_key = config.birdeye.api_key, chain = BirdeyeChain.ETHEREUM.value)
        token_address = WETH

        # load mock response
        mock_file_name = "get_token_overview_ethereum"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenOverviewResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = birdeye.client.get_token_overview(token_address)

        # actual test
        assert isinstance(response, GetTokenOverviewResponse)
        assert response.data.address == token_address

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_token_overview_ethereum_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Token - Overview" 
            specifically for Ethereum chain for asynchronous logic.

            Mock Response File: get_token_overview_ethereum.json
        """
        birdeye = Birdeye(api_key = config.birdeye.api_key, chain = BirdeyeChain.ETHEREUM.value)
        token_address = WETH

        # load mock response
        mock_file_name = "get_token_overview_ethereum"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTokenOverviewResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with birdeye.async_client as client:
            response = await client.get_token_overview(token_address)

        # actual test
        assert isinstance(response, GetTokenOverviewResponse)
        assert response.data.address == token_address

    def test_get_trades_token_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Trades - Token" 
            for synchronous logic.

            Mock Response File: get_trades_token.json
        """

        # load mock response
        mock_file_name = "get_trades_token"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTradesTokenResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_trades_token(SOL)

        # actual test
        assert isinstance(response, GetTradesTokenResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_trades_token_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Trades - Token" 
            for asynchronous logic.

            Mock Response File: get_trades_token.json
        """

        # load mock response
        mock_file_name = "get_trades_token"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTradesTokenResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_trades_token(SOL)

        # actual test
        assert isinstance(response, GetTradesTokenResponse)

    def test_get_trades_pair_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Trades - Pair" 
            for synchronous logic.

            Mock Response File: get_trades_pair.json
        """
        # load mock response
        mock_file_name = "get_trades_pair"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTradesPairResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_trades_pair(address = TOM_SOL)

        # actual test
        assert isinstance(response, GetTradesPairResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_trades_pair_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Trades - Pair" 
            for asynchronous logic.

            Mock Response File: get_trades_pair.json
        """
        # load mock response
        mock_file_name = "get_trades_pair"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetTradesPairResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_trades_pair(address = TOM_SOL)

        # actual test
        assert isinstance(response, GetTradesPairResponse)

    def test_get_ohlcv_incorrect_input_dates_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the incorrect dates inputs (dt_from > dt_to) 
            for synchronous logic.
        """

        with pytest.raises(BirdeyeTimeRangeError):
            # execute request
            self.birdeye.client.get_ohlcv(
                address = SOL,
                address_type = BirdeyeAddressType.TOKEN.value,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() + timedelta(hours = 1)
            )

    @pytest.mark.asyncio
    async def test_get_ohlcv_incorrect_input_dates_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the incorrect dates inputs (dt_from > dt_to) 
            for asynchronous logic.
        """

        with pytest.raises(BirdeyeTimeRangeError):
            # execute request
            async with self.birdeye.async_client as client:
                await client.get_ohlcv(
                    address = SOL,
                    address_type = BirdeyeAddressType.TOKEN.value,
                    timeframe = BirdeyeTimeFrame.MIN15.value,
                    dt_from = datetime.now() + timedelta(hours = 1)
                )

    def test_get_ohlcv_token_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "OHLCV - Token" 
            for synchronous logic.

            Mock Response File: get_ohlcv_token.json
        """

        # load mock response
        mock_file_name = "get_ohlcv_token"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetOHLCVTokenPairResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_ohlcv(
            address = SOL,
            address_type = BirdeyeAddressType.TOKEN.value,
            timeframe = BirdeyeTimeFrame.MIN15.value,
            dt_from = datetime.now() - timedelta(hours = 1),
            dt_to = datetime.now()
        )

        # actual test
        assert isinstance(response, GetOHLCVTokenPairResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_ohlcv_token_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "OHLCV - Token" 
            for asynchronous logic.

            Mock Response File: get_ohlcv_token.json
        """

        # load mock response
        mock_file_name = "get_ohlcv_token"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetOHLCVTokenPairResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_ohlcv(
                address = SOL,
                address_type = BirdeyeAddressType.TOKEN.value,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() - timedelta(hours = 1),
                dt_to = datetime.now()
            )

        # actual test
        assert isinstance(response, GetOHLCVTokenPairResponse)

    def test_get_ohlcv_pair_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "OHLCV - Pair" 
            for synchronous logic.

            Mock Response File: get_ohlcv_pair.json
        """

        # load mock response
        mock_file_name = "get_ohlcv_pair"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetOHLCVTokenPairResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_ohlcv(
            address = SOL,
            address_type = BirdeyeAddressType.PAIR.value,
            timeframe = BirdeyeTimeFrame.MIN15.value,
            dt_from = datetime.now() - timedelta(hours = 1),
            dt_to = datetime.now()
        )

        # actual test
        assert isinstance(response, GetOHLCVTokenPairResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_ohlcv_pair_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "OHLCV - Pair" 
            for asynchronous logic.

            Mock Response File: get_ohlcv_pair.json
        """

        # load mock response
        mock_file_name = "get_ohlcv_pair"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetOHLCVTokenPairResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_ohlcv(
                address = SOL,
                address_type = BirdeyeAddressType.PAIR.value,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() - timedelta(hours = 1),
                dt_to = datetime.now()
            )

        # actual test
        assert isinstance(response, GetOHLCVTokenPairResponse)

    def test_get_ohlcv_base_quote_incorrect_input_dates_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the incorrect dates inputs (dt_from > dt_to) 
            for synchronous logic.
        """

        with pytest.raises(BirdeyeTimeRangeError):
            # execute request
            self.birdeye.client.get_ohlcv_base_quote(
                base_address = SOL,
                quote_address = USDC,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() + timedelta(hours = 1)
            )

    @pytest.mark.asyncio
    async def test_get_ohlcv_base_quote_incorrect_input_dates_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the incorrect dates inputs (dt_from > dt_to) 
            for asynchronous logic.
        """

        with pytest.raises(BirdeyeTimeRangeError):
            # execute request
            async with self.birdeye.async_client as client:
                await client.get_ohlcv_base_quote(
                    base_address = SOL,
                    quote_address = USDC,
                    timeframe = BirdeyeTimeFrame.MIN15.value,
                    dt_from = datetime.now() + timedelta(hours = 1)
                )

    def test_get_ohlcv_base_quote_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "OHLCV - Base/Quote" 
            for synchronous logic.

            Mock Response File: get_ohlcv_base_quote.json
        """

        # load mock response
        mock_file_name = "get_ohlcv_base_quote"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetOHLCVBaseQuoteResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_ohlcv_base_quote(
            base_address = SOL,
            quote_address = USDC,
            timeframe = BirdeyeTimeFrame.MIN15.value,
            dt_from = datetime.now() - timedelta(hours = 1),
            dt_to = datetime.now()
        )

        # actual test
        assert isinstance(response, GetOHLCVBaseQuoteResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_ohlcv_base_quote_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "OHLCV - Base/Quote" 
            for asynchronous logic.

            Mock Response File: get_ohlcv_base_quote.json
        """

        # load mock response
        mock_file_name = "get_ohlcv_base_quote"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetOHLCVBaseQuoteResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_ohlcv_base_quote(
                base_address = SOL,
                quote_address = USDC,
                timeframe = BirdeyeTimeFrame.MIN15.value,
                dt_from = datetime.now() - timedelta(hours = 1),
                dt_to = datetime.now()
            )

        # actual test
        assert isinstance(response, GetOHLCVBaseQuoteResponse)

    def test_get_wallet_supported_networks_sync(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Wallet - Supported Networks" 
            for synchronous logic.

            Mock Response File: get_wallet_supported_networks.json
        """

        # load mock response
        mock_file_name = "get_wallet_supported_networks"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetWalletSupportedNetworksResponse)
            mocker.patch("cyhole.core.client.APIClient.api", return_value = mock_response)
            
        # execute request
        response = self.birdeye.client.get_wallet_supported_networks()

        # actual test
        assert isinstance(response, GetWalletSupportedNetworksResponse)

        # store request (only not mock)
        if config.mock_file_overwrite and not config.birdeye.mock_response_private:
            self.mocker.store_mock_model(mock_file_name, response)

    @pytest.mark.asyncio
    async def test_get_wallet_supported_networks_async(self, mocker: MockerFixture) -> None:
        """
            Unit Test used to check the response schema of endpoint "Wallet - Supported Networks" 
            for asynchronous logic.

            Mock Response File: get_wallet_supported_networks.json
        """

        # load mock response
        mock_file_name = "get_wallet_supported_networks"
        if config.mock_response or config.birdeye.mock_response_private:
            mock_response = self.mocker.load_mock_response(mock_file_name, GetWalletSupportedNetworksResponse)
            mocker.patch("cyhole.core.client.AsyncAPIClient.api", return_value = mock_response)
            
        # execute request
        async with self.birdeye.async_client as client:
            response = await client.get_wallet_supported_networks()

        # actual test
        assert isinstance(response, GetWalletSupportedNetworksResponse)


from ..core.param import CyholeParam

class BirdeyeChain(CyholeParam):
    """
        Enum class to get the supported Birdeye's chains.
    """
    SOLANA = "solana"
    """Identifier of the Solana chain in Birdeye API."""
    ETHEREUM = "ethereum"
    """Identifier of the Ethereum chain in Birdeye API."""
    ARBITRUM = "arbitrum"
    """Identifier of the Arbitrum chain in Birdeye API."""
    AVALANCHE = "avalanche"
    """Identifier of the Avalanche chain in Birdeye API."""
    BSC = "bsc"
    """Identifier of the BSC (Binance Smart Chain) chain in Birdeye API."""
    OPTIMISM = "optimism"
    """Identifier of the Optimism chain in Birdeye API."""
    POLYGON = "polygon"
    """Identifier of the Polygon chain in Birdeye API."""
    BASE = "base"
    """Identifier of the Base chain in Birdeye API."""
    ZKSYNC = "zksync"
    """Identifier of the zkSync chain in Birdeye API."""

class BirdeyeOrder(CyholeParam):
    """
        Enum class to get the supported Birdeye's orders.
    """
    ASCENDING = "asc"
    """Ascending sort order."""
    DESCENDING = "desc"
    """Descending sort order."""

class BirdeyeSort(CyholeParam):
    """
        Enum class to get the supported Birdeye's sorting types.
    """
    SORT_MC = "mc"
    """Sort the results by Market Cap (MC)"""
    SORT_V24HUSD = "v24hUSD"
    """Sort the results by USD Volume in the last 24h"""
    SORT_V24HCHANGEPERCENT = "v24hChangePercent"
    """Sort the results by percent changing in the last 24h"""

class BirdeyeTimeFrame(CyholeParam):
    """
        Enum class to get the supported Birdeye's timeframe sizes.
    """
    MIN1 = "1m"
    """1 minute"""
    MIN3 = "3m"
    """3 minute"""
    MIN5 = "5m"
    """5 minute"""
    MIN15 = "15m"
    """15 minute"""
    MIN30 = "30m"
    """30 minute"""
    H1 = "1H"
    """1 hour"""
    H2 = "2H"
    """2 hour"""
    H4 = "4H"
    """4 hour"""
    H6 = "6H"
    """6 hour"""
    H8 = "8H"
    """8 hour"""
    H12 = "12H"
    """12 hour"""
    D1 = "1D"
    """1 day"""
    D3 = "3D"
    """3 day"""
    W1 = "1W"
    """1 week"""
    M1 = "1M"
    """1 month"""

class BirdeyeAddressType(CyholeParam):
    """
        Enum class to get the supported Birdeye's address' types.
    """
    TOKEN = "token"
    """The address is refering to a token on the chain."""
    PAIR = "pair"
    """The address is refering to a token pair on the chain. (e.g. SOL/USDT)"""

class BirdeyeTradeType(CyholeParam):
    """
        Enum class to get the supported Birdeye's transactions' types.
    """
    SWAP = "swap"
    """A classic exchange between two currencies."""
    ADD = "add"
    """Transactions that add liquidity."""
    REMOVE = "remove"
    """Transactions that remove liquidity."""
    ALL = "all"
    """All type of transactions (swap, add, remove)."""


import pytest

from cyhole.core.interaction import Interaction
from cyhole.core.client import APIClient, AsyncAPIClient
from cyhole.core.param import CyholeParam, RequestType
from cyhole.core.exception import RequestTypeNotSupported, ParamUnknownError, AsyncClientAPISessionNotAvailable

URL_TEST_GET = "http://httpbin.org/get"
URL_TEST_POST = "http://httpbin.org/post"

# variables
interaction = Interaction()
client = APIClient(interaction)

def test_sync_client_api_request_type_not_supported() -> None:
    """
        Unit Test for `APIClient.api` function with Request Type not supported.
    """
    with pytest.raises(RequestTypeNotSupported):
        client.api(type = "XXX", url = "")

def test_sync_client_api_with_header() -> None:
    """
        Unit Test for `APIClient.api` function with `headers`.
    """
    with pytest.raises(RequestTypeNotSupported):
        client.api(type = "XXX", url = "", headers = {"test": "test"})

def test_sync_client_api_get() -> None:
    """
        Unit Test for `APIClient.api` function for GET endpoint.
    """
    response = client.api(type = RequestType.GET.value, url = URL_TEST_GET)
    assert response.status_code == 200
    assert response.content.decode() is not None

def test_sync_client_api_post() -> None:
    """
        Unit Test for `APIClient.api` function for POST endpoint.
    """
    headers = {
        "Content-Type": "application/json"
    }
    interaction = Interaction(headers = headers)
    client = APIClient(interaction)
    response = client.api(
        type = RequestType.POST.value,
        url = URL_TEST_POST,
        json = {"test": "data"}
    )
    assert response.status_code == 200
    assert response.content.decode() is not None

@pytest.mark.asyncio
async def test_async_client_init() -> None:
    """
        Unit Test to check the correct init of AsyncAPIClient.
    """
    async_client = AsyncAPIClient(interaction)

    assert async_client._session is None
    assert not async_client.is_connected()

@pytest.mark.asyncio
async def test_async_client_connect() -> None:
    """
        Unit Test to check the correct connection of AsyncAPIClient.
    """
    async_client = AsyncAPIClient(interaction)
    async_client.connect()

    assert async_client._session is not None
    assert async_client.is_connected()

@pytest.mark.asyncio
async def test_async_client_close_connetion() -> None:
    """
        Unit Test to check the correct closing connection of AsyncAPIClient.
    """
    async_client = AsyncAPIClient(interaction)
    async_client.connect()
    await async_client.close()

    assert async_client._session is None
    assert not async_client.is_connected()

@pytest.mark.asyncio
async def test_async_client_close_connetion_error() -> None:
    """
        Unit Test to check error on closing connection 
        if session not available on AsyncAPIClient.
    """
    async_client = AsyncAPIClient(interaction)
    with pytest.raises(AsyncClientAPISessionNotAvailable):
        await async_client.close()

@pytest.mark.asyncio
async def test_async_client_context_manager() -> None:
    """
        Unit Test to check the correct usage of context manager of AsyncAPIClient.
    """
    async with AsyncAPIClient(interaction) as client:
        assert client._session is not None
        assert client.is_connected()

@pytest.mark.asyncio
async def test_async_client_api_request_no_session() -> None:
    """
        Unit Test for `AsyncAPIClient.api` function with Request Type not supported.
    """
    async_client = AsyncAPIClient(interaction)
    with pytest.raises(AsyncClientAPISessionNotAvailable):
        await async_client.api(type = RequestType.GET.value, url = URL_TEST_GET)

@pytest.mark.asyncio
async def test_async_client_api_request_type_not_supported() -> None:
    """
        Unit Test for `AsyncAPIClient.api` function with Request Type not supported.
    """
    async with AsyncAPIClient(interaction) as client:
        with pytest.raises(RequestTypeNotSupported):
            await client.api(type = "XXX", url = "")

@pytest.mark.asyncio
async def test_async_client_api_get() -> None:
    """
        Unit Test for `AsyncAPIClient.api` function for GET endpoint.
    """
    async with AsyncAPIClient(interaction) as client:
        response = await client.api(type = RequestType.GET.value, url = URL_TEST_GET)
        assert response.status_code == 200
        assert response.content.decode() is not None

@pytest.mark.asyncio
async def test_async_client_api_get_with_params() -> None:
    """
        Unit Test for `AsyncAPIClient.api` function for GET endpoint.
    """
    params = {
        "name": "cyhole",
        "version" : None
    }
    async with AsyncAPIClient(interaction) as client:
        response = await client.api(type = RequestType.GET.value, url = URL_TEST_GET, params = params)
        assert response.status_code == 200
        assert response.content.decode() is not None

@pytest.mark.asyncio
async def test_async_client_api_post() -> None:
    """
        Unit Test for `AsyncAPIClient.api` function for POST endpoint.
    """
    headers = {
        "Content-Type": "application/json"
    }
    interaction = Interaction(headers = headers)
    async with AsyncAPIClient(interaction) as client:
        response = await client.api(
            type = RequestType.POST.value,
            url = URL_TEST_POST,
            json = {"test": "data"})
        assert response.status_code == 200
        assert response.content.decode() is not None

def test_param_unknown() -> None:
    """
        Unit Test for `ParamUnknownError` exception.
    """
    class ParamTest(CyholeParam):
        TEST = "test"

    with pytest.raises(ParamUnknownError):
        ParamTest.check("xxx")


from typing import Type, Any
from enum import Enum

class CyholeException(Exception):
    """
        General exception of Cyhole library
    """
    def __init__(self, description):
        self.description = description

    def __str__(self):
        return self.description

class RequestTypeNotSupported(CyholeException):
    pass

class AuthorizationAPIKeyError(CyholeException):
    pass

class MissingAPIKeyError(CyholeException):
    pass

class ParamUnknownError(CyholeException):
    """
        General error used to manage a wrong input API param.
    """
    def __init__(self, param_value: Any, param_enum: Type[Enum]):
        description = f"param '{str(param_value)}' not supported in {param_enum.__name__} enum class. \nAdmissible values: {[(param.name, param.value) for param in param_enum]}"
        super().__init__(description)

class AsyncClientAPISessionNotAvailable(CyholeException):
    """
        Exception raised when the [`AsyncAPIClient`][cyhole.core.AsyncAPIClient] has not an open session to use.
    """
    pass


from requests.exceptions import HTTPError
from typing import Any, Coroutine, overload, Literal

from ..core.param import RequestType
from ..core.interaction import Interaction
from ..jupiter.client import JupiterClient, JupiterAsyncClient
from ..jupiter.schema import (
    JupiterHTTPError,
    GetPriceResponse,
    GetQuoteInput,
    GetQuoteResponse,
    GetQuoteTokensResponse,
    GetQuoteProgramIdLabelResponse,
    PostSwapBody,
    PostSwapResponse,
    GetTokenListResponse,
    PostLimitOrderCreateBody,
    PostLimitOrderCreateResponse,
    PostLimitOrderCancelBody,
    PostLimitOrderCancelResponse,
    GetLimitOrderOpenResponse,
    GetLimitOrderHistoryResponse,
    GetLimitOrderTradeHistoryResponse
)
from ..jupiter.exception import (
    JupiterException,
    JupiterNoRouteFoundError,
    JupiterInvalidRequest
)
from ..jupiter.param import JupiterTokenListType

class Jupiter(Interaction):
    """
        Class used to connect [Jupiter](https://jup.ag) API, one of them most popular Solana DEX. 
        To have access Jupiter API is **not** required an API key.

        Check [https://station.jup.ag/docs/api](https://station.jup.ag/docs/api) for all the details on the available endpoints.

        **Example**

        ```python
        import asyncio
        from cyhole.jupiter import Jupiter
        from cyhole.core.address.solana import JUP

        jupiter = Jupiter()

        # Get current price of JUP on Solana
        # synchronous
        response = jupiter.client.get_price([JUP])
        print("Current JUP/USDC:", response.data[JUP].price)

        # asynchronous
        async def main() -> None:
            async with jupiter.async_client as client:
                response = await client.get_price([JUP])
                print("Current JUP/USDC:", response.data[JUP].price)

        asyncio.run(main())
        ```
    """

    def __init__(self, headers: Any | None = None) -> None:
        super().__init__(headers)

        # clients
        self.client = JupiterClient(self)
        self.async_client = JupiterAsyncClient(self)

        # API urls
        self.url_api_price = "https://price.jup.ag/v6/price"
        self.url_api_quote = "https://quote-api.jup.ag/v6/"
        self.url_api_token = "https://token.jup.ag/"
        self.url_api_limit = "https://jup.ag/api/limit/v1/"
        return

    @overload
    def _get_price(self, sync: Literal[True], address: list[str], vs_address: str | None = None) -> GetPriceResponse: ...

    @overload
    def _get_price(self, sync: Literal[False], address: list[str], vs_address: str | None = None) -> Coroutine[None, None, GetPriceResponse]: ...

    def _get_price(self, sync: bool, address: list[str], vs_address: str | None = None) -> GetPriceResponse | Coroutine[None, None, GetPriceResponse]:
        """
            This function refers to the **[Price](https://station.jup.ag/docs/apis/price-api)** API endpoint, 
            and it is used to get the current price of a list of tokens on Solana chain with respect to another token. 

            The API returns the unit buy price for the tokens; by default, the price is provided according to 
            the value of `USDC` token. It is also possible to provide another comparison token in the request.

            !!! info
                Observe that when the token address/symbol or comparison token address/symbol are not found, 
                the response provided will have an empty `data` object.

            Parameters:
                address: list of tokens involved in the request.
                    It is possible to provide the Solana tokens' addresses or their symbols. 
                    For example, `SOL` is equivalent to `So11111111111111111111111111111111111111112`.
                vs_address: comparison token address/symbol.
                    Default Value: `None` (`USDC`)

            Returns:
                tokens' prices.
        """
        # set params
        params = {
            "ids": ",".join(address),
            "vsToken": vs_address
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, self.url_api_price, params = params)
            return GetPriceResponse(**content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, self.url_api_price, params = params)
                return GetPriceResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_quote(self, sync: Literal[True], input: GetQuoteInput) -> GetQuoteResponse: ...

    @overload
    def _get_quote(self, sync: Literal[False], input: GetQuoteInput) -> Coroutine[None, None, GetQuoteResponse]: ...

    def _get_quote(self, sync: bool, input: GetQuoteInput) -> GetQuoteResponse | Coroutine[None, None, GetQuoteResponse]:
        """
            This function refers to the **[Get Quote](https://station.jup.ag/api-v6/get-quote)** API endpoint, 
            and it is used to get a quote for swapping a specific amount of tokens.  
            The function can be combined with the `post_swap` to implement a payment mechanism.

            Parameters:
                input: an input schema used to describe the request.
                    More details in the object definition.

            Returns:
                Quote found by Jupiter API.
        """
        # set params
        url = self.url_api_quote + "quote"
        params = input.model_dump(
            by_alias = True, 
            exclude_defaults = True
        )

        # execute request
        if sync:
            try:
                content_raw = self.client.api(RequestType.GET.value, url, params = params)
            except HTTPError as e:
                raise self._raise(e)
            return GetQuoteResponse(**content_raw.json())
        else:
            async def async_request():
                try:
                    content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                except HTTPError as e:
                    raise self._raise(e)
                return GetQuoteResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_quote_tokens(self, sync: Literal[True]) -> GetQuoteTokensResponse: ...

    @overload
    def _get_quote_tokens(self, sync: Literal[False]) -> Coroutine[None, None, GetQuoteTokensResponse]: ...

    def _get_quote_tokens(self, sync: bool) -> GetQuoteTokensResponse | Coroutine[None, None, GetQuoteTokensResponse]:
        """
            This function refers to the **[Get Quote Tokens](https://station.jup.ag/api-v6/get-tokens)** API endpoint, 
            and it is used to get the list of the current supported tradable tokens. 

            Returns:
                List of tradable tokens.
        """
        # set params
        url = self.url_api_quote + "tokens"

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url)
            return GetQuoteTokensResponse(tokens = content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url)
                return GetQuoteTokensResponse(tokens = content_raw.json())
            return async_request()

    @overload
    def _get_quote_program_id_label(self, sync: Literal[True]) -> GetQuoteProgramIdLabelResponse: ...

    @overload
    def _get_quote_program_id_label(self, sync: Literal[False]) -> Coroutine[None, None, GetQuoteProgramIdLabelResponse]: ...

    def _get_quote_program_id_label(self, sync: bool) -> GetQuoteProgramIdLabelResponse | Coroutine[None, None, GetQuoteProgramIdLabelResponse]:
        """
            This function refers to the **[Get Quote Program ID to Label](https://station.jup.ag/api-v6/get-program-id-to-label)** API endpoint, 
            and it is used to get the list of supported DEXes to use in quote endpoint. 

            Returns:
                List of DEXs addresses and labels.
        """
        # set params
        url = self.url_api_quote + "program-id-to-label"

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url)
            return GetQuoteProgramIdLabelResponse(dexes = content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url)
                return GetQuoteProgramIdLabelResponse(dexes = content_raw.json())
            return async_request()

    @overload
    def _post_swap(self, sync: Literal[True], body: PostSwapBody) -> PostSwapResponse: ...

    @overload
    def _post_swap(self, sync: Literal[False], body: PostSwapBody) -> Coroutine[None, None, PostSwapResponse]: ...

    def _post_swap(self, sync: bool, body: PostSwapBody) -> PostSwapResponse | Coroutine[None, None, PostSwapResponse]:
        """
            This function refers to the **[Post Swap](https://station.jup.ag/api-v6/post-swap)** API endpoint, 
            and it is used to recive the transaction to perform the swap initialised from `get_quote` 
            endopoint for the desired pair.  
            The function should be combined with the `get_quote` endpoint.

            Parameters:
                body: the body to sent to Jupiter API that describe the swap.
                    More details in the object definition.

            Returns:
                Transaction found by Jupiter API.
        """
        # set params
        url = self.url_api_quote + "swap"
        headers = {
            "Content-Type": "application/json"
        }

        # execute request
        if sync:
            try:
                content_raw = self.client.api(
                    type = RequestType.POST.value,
                    url = url,
                    headers = headers,
                    json = body.model_dump(by_alias = True, exclude_defaults = True)
                )
            except HTTPError as e:
                raise self._raise(e)
            return PostSwapResponse(**content_raw.json())
        else:
            async def async_request():
                try:
                    content_raw = await self.async_client.api(
                        type = RequestType.POST.value,
                        url = url,
                        headers = headers,
                        json = body.model_dump(by_alias = True, exclude_defaults = True)
                    )
                except HTTPError as e:
                    raise self._raise(e)
                return PostSwapResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_token_list(self, sync: Literal[True], type: str = JupiterTokenListType.STRICT.value, banned: None | bool = None) -> GetTokenListResponse: ...

    @overload
    def _get_token_list(self, sync: Literal[False], type: str = JupiterTokenListType.STRICT.value, banned: None | bool = None) -> Coroutine[None, None, GetTokenListResponse]: ...

    def _get_token_list(self, sync: bool, type: str = JupiterTokenListType.STRICT.value, banned: None | bool = None) -> GetTokenListResponse | Coroutine[None, None, GetTokenListResponse]:
        """
            This function refers to the **[Token List](https://station.jup.ag/docs/token-list/token-list-api)** API endpoint, 
            and it is used to retrieved the list of tokens eligible for trading, managed by Jupiter.  
            Choose the tokens list according to `type` field.

            Parameters:
                type: Jupiter manages the tradable tokens through a set of tags in order to guarantee its 
                    core values and provide a secure service. There are two types of lists currently available:  
                    - Strict the most secure list that includes only the tokens with tags `old-registry`, `community`, or `wormhole`.
                    - All all the tokens are included expect the banned ones.
                    The supported types are available on [`JupiterTokenListType`][cyhole.jupiter.param.JupiterTokenListType].
                banned: this flag can be used **only** on `All` type, and it enables the inclusion of banned tokens.

            Returns:
                List of Jupiter's tokens list.
        """
        # check param consistency
        JupiterTokenListType.check(type)

        # set params
        url = self.url_api_token + type
        params = {
            "includeBanned" : banned
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetTokenListResponse(tokens = content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetTokenListResponse(tokens = content_raw.json())
            return async_request()

    @overload
    def _post_limit_order_create(self, sync: Literal[True], body: PostLimitOrderCreateBody) -> PostLimitOrderCreateResponse: ...

    @overload
    def _post_limit_order_create(self, sync: Literal[False], body: PostLimitOrderCreateBody) -> Coroutine[None, None, PostLimitOrderCreateResponse]: ...

    def _post_limit_order_create(self, sync: bool, body: PostLimitOrderCreateBody) -> PostLimitOrderCreateResponse | Coroutine[None, None, PostLimitOrderCreateResponse]:
        """
            This function refers to the **[Post Limit Order - Create](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint, 
            and it is used to receive the transaction to perform the creation of a Limit Order via Jupiter API.

            Parameters:
                body: the body to sent to Jupiter API that describe the swap.
                    More details in the object definition.

            Returns:
                Transaction created by Jupiter API.
        """

        # set params
        url = self.url_api_limit + "createOrder"
        headers = {
            "Content-Type": "application/json"
        }

        # execute request
        if sync:
            try:
                content_raw = self.client.api(
                    type = RequestType.POST.value,
                    url = url,
                    headers = headers,
                    json = body.model_dump(by_alias = True, exclude_defaults = True)
                )
            except HTTPError as e:
                raise self._raise(e)
            return PostLimitOrderCreateResponse(**content_raw.json())
        else:
            async def async_request():
                try:
                    content_raw = await self.async_client.api(
                        type = RequestType.POST.value,
                        url = url,
                        headers = headers,
                        json = body.model_dump(by_alias = True, exclude_defaults = True)
                    )
                except HTTPError as e:
                    raise self._raise(e)
                return PostLimitOrderCreateResponse(**content_raw.json())
            return async_request()

    @overload
    def _post_limit_order_cancel(self, sync: Literal[True], body: PostLimitOrderCancelBody) -> PostLimitOrderCancelResponse: ...

    @overload
    def _post_limit_order_cancel(self, sync: Literal[False], body: PostLimitOrderCancelBody) -> Coroutine[None, None, PostLimitOrderCancelResponse]: ...

    def _post_limit_order_cancel(self, sync: bool, body: PostLimitOrderCancelBody) -> PostLimitOrderCancelResponse | Coroutine[None, None, PostLimitOrderCancelResponse]:
        """
            This function refers to the **[Post Limit Order - Cancel](https://station.jup.ag/docs/limit-order/limit-order-api)** API endpoint, 
            and it is used to receive the transaction to perform the cancellation of a Limit Order previously 
            opened via Jupiter API.

            Parameters:
                body: the body to sent to Jupiter API that describe the limit order to cancel.
                    More details in the object definition.

            Returns:
                Transaction created by Jupiter API.
        """

        # set params
        url = self.url_api_limit + "cancelOrders"
        headers = {
            "Content-Type": "application/json"
        }

        # execute request
        if sync:
            try:
                content_raw = self.client.api(
                    type = RequestType.POST.value,
                    url = url,
                    headers = headers,
                    json = body.model_dump(by_alias = True, exclude_defaults = True)
                )
            except HTTPError as e:
                raise self._raise(e)
            return PostLimitOrderCancelResponse(**content_raw.json())
        else:
            async def async_request():
                try:
                    content_raw = await self.async_client.api(
                        type = RequestType.POST.value,
                        url = url,
                        headers = headers,
                        json = body.model_dump(by_alias = True, exclude_defaults = True)
                    )
                except HTTPError as e:
                    raise self._raise(e)
                return PostLimitOrderCancelResponse(**content_raw.json())
            return async_request()

    @overload
    def _get_limit_order_open(
            self,
            sync: Literal[True],
            wallet: str | None = None,
            input_token: str | None = None,
            output_token: str | None = None
        ) -> GetLimitOrderOpenResponse: ...

    @overload
    def _get_limit_order_open(
            self,
            sync: Literal[False],
            wallet: str | None = None,
            input_token: str | None = None,
            output_token: str | None = None
        ) -> Coroutine[None, None, GetLimitOrderOpenResponse]: ...

    def _get_limit_order_open(
        self,
        sync: bool,
        wallet: str | None = None,
        input_token: str | None = None,
        output_token: str | None = None
    ) -> GetLimitOrderOpenResponse | Coroutine[None, None, GetLimitOrderOpenResponse]:
        """
            This function refers to the **[Get Limit Order - Open](https://station.jup.ag/docs/limit-order/limit-order-api)** 
            API endpoint, and it is used to receive the current open limit orders related to a wallet, input token 
            or output token via Jupiter API. 

            Observe that all the input parameters are optional; if for example, only the `input_token` is provided, 
            then **all** the limit orders having that input token address are returned (if available).

            Parameters:
                wallet: address of the wallet to check.
                input_token: address of the input token associated to the limit order.
                output_token: address of the output token associated to the limit order.

            Returns:
                Open limit orders created by Jupiter API.
        """
        # set params
        url = self.url_api_limit + "openOrders"
        params = {
            "wallet": wallet,
            "inputMint": input_token,
            "outputMint": output_token
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetLimitOrderOpenResponse(orders = content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetLimitOrderOpenResponse(orders = content_raw.json())
            return async_request()

    @overload
    def _get_limit_order_history(
        self,
        sync: Literal[True],
        wallet: str,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderHistoryResponse: ...

    @overload
    def _get_limit_order_history(
        self,
        sync: Literal[False],
        wallet: str,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> Coroutine[None, None, GetLimitOrderHistoryResponse]: ...

    def _get_limit_order_history(
        self,
        sync: bool,
        wallet: str,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderHistoryResponse | Coroutine[None, None, GetLimitOrderHistoryResponse]:
        """
            This function refers to the **[Get Limit Order - History](https://station.jup.ag/docs/limit-order/limit-order-api)** 
            API endpoint, and it is used to retrieve the history of Limit Orders associated to a wallet via Jupiter API. 

            Parameters:
                wallet: address of the wallet to check.
                cursor: specify which 'page' of orders to return.
                skip: specify the number of order to skip (from the top).
                take: specify the number of orders to return.

            Returns:
                History of limit orders associated to the input wallet.
        """
        # set params
        url = self.url_api_limit + "orderHistory"
        params = {
            "wallet": wallet,
            "cursor": cursor,
            "skip": skip,
            "take": take
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetLimitOrderHistoryResponse(orders = content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetLimitOrderHistoryResponse(orders = content_raw.json())
            return async_request()

    @overload
    def _get_limit_order_trade_history(
        self,
        sync: Literal[True],
        wallet: str | None = None,
        input_token: str | None = None,
        output_token: str | None = None,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderTradeHistoryResponse: ...

    @overload
    def _get_limit_order_trade_history(
        self,
        sync: Literal[False],
        wallet: str | None = None,
        input_token: str | None = None,
        output_token: str | None = None,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> Coroutine[None, None, GetLimitOrderTradeHistoryResponse]: ...

    def _get_limit_order_trade_history(
        self,
        sync: bool,
        wallet: str | None = None,
        input_token: str | None = None,
        output_token: str | None = None,
        cursor: int | None = None,
        skip: int | None = None,
        take: int | None = None
    ) -> GetLimitOrderTradeHistoryResponse | Coroutine[None, None, GetLimitOrderTradeHistoryResponse]:
        """
            This function refers to the **[Get Limit Order - Trade History](https://station.jup.ag/docs/limit-order/limit-order-api)** 
            API endpoint, and it is used to retrieve the trades history related to Limit Orders extracted with specific 
            requirements via Jupiter API. 

            Parameters:
                wallet: address of the wallet to check.
                input_token: address of the input token associated to the limit order.
                output_token: address of the output token associated to the limit order.
                cursor: specify which 'page' of orders to return.
                skip: specify the number of order to skip (from the top).
                take: specify the number of orders to return.

            Returns:
                Hostory of limit orders associated to the input wallet.
        """
        # set params
        url = self.url_api_limit + "tradeHistory"
        params = {
            "wallet": wallet,
            "inputMint": input_token,
            "outputMint": output_token,
            "cursor": cursor,
            "skip": skip,
            "take": take
        }

        # execute request
        if sync:
            content_raw = self.client.api(RequestType.GET.value, url, params = params)
            return GetLimitOrderTradeHistoryResponse(orders = content_raw.json())
        else:
            async def async_request():
                content_raw = await self.async_client.api(RequestType.GET.value, url, params = params)
                return GetLimitOrderTradeHistoryResponse(orders = content_raw.json())
            return async_request()

    def _raise(self, exception: HTTPError) -> JupiterException:
        """
            Internal function used to raise the correct 
            Jupiter exception according to the error code 
            provided by the API.

            Parameters:
                exception: the HTTP error returned from Jupiter API.

            Raises:
                JupiterNoRouteFoundError: for error code `COULD_NOT_FIND_ANY_ROUTE` 
                    during the creation of a quote.
                JupiterInvalidRequest: for error code `INVALID_REQUEST`.
                JupiterException: general exception raised when an unknown 
                    error code is found or a different error is found.
        """
        try:
            error = JupiterHTTPError(**exception.response.json())
            match error.code:
                case "COULD_NOT_FIND_ANY_ROUTE":
                    return JupiterNoRouteFoundError(error.msg)
                case "INVALID_REQUEST":
                    return JupiterInvalidRequest(error.msg)
                case _:
                    return JupiterException(error.model_dump_json())
        except Exception:
            return JupiterException(exception.response.content.decode())